#include <assert.h>
#include "MISparseVector.h"
#include <math.h>
#include <stdio.h>
#include <string.h>

MISparseVector::MISparseVector()
  : label_(0),
    size_(0)
{
  strcpy(imgname,"");
}

MISparseVector::~MISparseVector()
{
  instances_.clear();
  free(pairs);
}

lasvm_sparsevector_t* MISparseVector::getInstance(int i)
{
  assert(i < size_);
  return instances_[i];
}

void MISparseVector::getRepresentation(double * out){
  lasvm_sparsevector_t * res = lasvm_sparsevector_create();
  for(unsigned int i=0;i<instances_.size();i++){
    lasvm_sparsevector_t * tmp = lasvm_sparsevector_combine(res,1,instances_[i],1);
    lasvm_sparsevector_destroy(res);
    res = tmp;
  }
  lasvm_sparsevector_pair_t *p1 = res->pairs;
  while(p1->index>=0){
    out[p1->index-1] = p1->value/size_;
    p1++;
  }
  lasvm_sparsevector_destroy(res);
}

lasvm_sparsevector_t * MISparseVector::getRepresentation(){
  lasvm_sparsevector_t * res = lasvm_sparsevector_create();
  for(unsigned int i=0;i<instances_.size();i++){
    lasvm_sparsevector_t * tmp = lasvm_sparsevector_combine(res,1,instances_[i],probs[i]/size_);
    lasvm_sparsevector_destroy(res);
    res = tmp;
  }
  return res;
}


void MISparseVector::addInstance(lasvm_sparsevector_t* instance)
{
  instances_.push_back(instance);
  probs.push_back(1);
  size_++;
}

void MISparseVector::deleteInstance(int pos)
{
  printf("Deleting instance %d\n",pos);
  if(size_>0){
    instances_.erase(instances_.begin()+pos);
    size_--;
  }
}

void MISparseVector::deletePointer(lasvm_sparsevector_t * p){
  for(int i=0;i<size_;i++)
    if(p==instances_[i]){
      deleteInstance(i);
      break;
    }
}

int MISparseVector::size() const
{
  return size_;
}

void MISparseVector::write(ofstream &stream_out)
{
  stream_out << size_ << endl;
  for (int i = 0 ; i < size_; i++)
    instances_[i]->write(stream_out);
}

void MISparseVector::printX(int dim){
  double out[dim];
  for(int i=0;i<dim;i++)
    out[i] = 0;
  getRepresentation(out);
  for(int i=0;i<dim;i++)
    //if(out[i]!=0)
    printf("%d:%lf ",i+1,out[i]);
  printf("\n");
}

double MISparseVector::norm() const
{
  return size_;
}

double MISparseVector::dot_product(MISparseVector *misv) const
{
  double dot = 0.0;
  for (int i = 0; i < size_; i++) {
    for (int j = 0; j < misv->size_; j++){
      dot += lasvm_sparsevector_dot_product(instances_[i], misv->instances_[j]) * probs[i] * misv->probs[j];
      /*if(bag_id == 2 && misv->bag_id == 1427){
	printf("dot = %lf\n",dot);
	printf("%lf %lf\n",probs[i],misv->probs[j]);
	}*/
    }
  }

  return dot;
}


double MISparseVector::rbf_dot_product(MISparseVector *misv,double gamma) const
{
  double dot = 0.0;
  double sq1[size_];
  double sq2[misv->size_];

  for (int i = 0; i < size_; i++) {
    sq1[i] = lasvm_sparsevector_dot_product(instances_[i], instances_[i]);
  }
  
  for (int j = 0; j < misv->size_; j++){
    sq2[j] = lasvm_sparsevector_dot_product(misv->instances_[j], misv->instances_[j]);
  }
  
  for (int i = 0; i < size_; i++) {
    for (int j = 0; j < misv->size_; j++){
      double dot_prod = lasvm_sparsevector_dot_product(instances_[i], misv->instances_[j]);
      //printf("%lf %lf %lf %lf\n",sq1[i]+sq2[i]-2*dot_prod,gamma,-gamma*(sq1[i]+sq2[j]-2*dot_prod),exp(-gamma*(sq1[i]+sq2[j]-2*dot_prod)));
      dot += exp(-gamma*(sq1[i]+sq2[j]-2*dot_prod));
    }
  }
  return dot;
}
