// -*- Mode: c++; c-file-style: "stroustrup"; -*-

using namespace std;

#include <stdio.h>
#include <vector>
#include <math.h>
#include <assert.h>

#include <ctime>
#include <iostream>
#include <fstream>
#include <string>
#include <limits>
#include <set>
#include <map>
#include <algorithm>

#include "libincsvm.h"
#include "libvoi.h"
#include "vector.h"
#include "MISparseVector.h"
#include "strutil.h"
#include "LabelConfidence.h"
#include "svqp2.h"

#define LINEAR  0
#define POLY    1
#define RBF     2
#define SIGMOID 3 

#define ONLINE 0
#define ONLINE_WITH_FINISHING 1

#define RANDOM 0
#define GRADIENT 1
#define MARGIN 2

#define ITERATIONS 0
#define SVS 1
#define TIME 2

#define TRAIN 0
#define TEST  1
#define UNIVERSUM 2
#define UNLABELED 3
#define SV 4
#define POSINSTANCES 5
#define UNLABELEDBAGS 6
#define TRAINPOS 7
#define CROSSVAL 8
#define TMP 9

#define SEQUENTIAL 10

#include <sys/resource.h>
#include <sys/time.h>
#include <unistd.h>
struct rlimit mylimit;



char *kernel_type_table[] = {"linear","polynomial","rbf","sigmoid"};
extern vector<int> ind[4];
extern vector<double> a;
extern double b;
int verbose = 0;

class ID // class to hold split file indices and labels
{
public:
    int x;
    int y;
    ID() : x(0), y(0) {}
    ID(int x1,int y1) : x(x1), y(y1) {}
};
// IDs will be sorted by index, not by label.
bool operator<(const ID& x, const ID& y)
{
    return x.x < y.x;
}

int initial_set_size = 1;
/* Data and model */
int m=0;                          // training set size
vector <lasvm_sparsevector_t*> X; // feature vectors
vector <int> Y;                   // labels
vector <double> kparam;           // kernel parameters
vector <double> alpha;            // alpha_i, SV weights
double b0;                        // threshold
vector <int> data_map[10]; // = new vector<int>[10];         // maps to the indices of train/test/... data
vector<MISparseVector *> bag_pointer;
int active_iterations = 1;
vector<double> Xsquare;
vector<int> got_label;
int mi_selection_type = ACTIVE_ALL;
double threshold = 0;
char model_file_name[1024];

vector<int> classes;
int rev_idx[50];
vector<int> train_split;
vector<int> unlab_split;

/* Hyperparameters */
int kernel_type=RBF;              // LINEAR, POLY, RBF or SIGMOID kernels
double degree=3,kgamma=-1,coef0=0;// kernel params
int use_b0=1;                     // use threshold via constraint \sum a_i y_i =0
int selection_type=SEQUENTIAL;        // RANDOM, GRADIENT or MARGIN selection strategies
int optimizer=ONLINE_WITH_FINISHING; // strategy of optimization
double C=1;                       // C, penalty on errors
double C_neg=1;                   // C-Weighting for negative examples
double C_pos=1;                   // C-Weighting for positive examples
int epochs=1;                     // epochs of online learning
int candidates=50;				  // number of candidates for "active" selection process
double deltamax=1000;			  // tolerance for performing reprocess step, 1000=1 reprocess only
vector <double> select_size;      // Max number of SVs to take with selection strategy (for early stopping) 
vector <double> x_square;         // norms of input vectors, used for RBF
double sigmoidA,sigmoidB;
double maxa=0;						// largest value of alpha


/* Programm behaviour*/
int verbosity=1;                  // verbosity level, 0=off
int saves=1;
char report_file_name[1024];             // filename for the training report
char testset_file_name[1024];             // filename for the training report
char labels_file_name[200]="none";
char fval_file_name[1024]="\0";
char split_file_name[1024]="\0";         // filename for the splits
char mi_un_split_file_[200]="none";         // filename for the splits
char test_file_name[200]="none";
char crossval_file_name[200]="none";
char class_labels[200]="\0";
int cache_size=256;                       // 256Mb cache size as default
double epsgr=1e-3;                       // tolerance on gradients
long long kcalcs=0;                      // number of kernel evaluations
int binary_files=0;
vector <ID> splits;             
int max_index=0;
vector <int> iold, inew;		  // sets of old (already seen) points + new (unseen) points
int termination_type=0;
char mi_split_file_[1024];
bool justtest = 0;                       // if justtest, then universvm tests on the specified model
double r12=1,r21=1;
double r_tradeoff;
int cost_of_labeling = 1;
int sample_size = 75;
int pos_bags = 0;
int sample_start = 50;
double total_cost_of_labeling = 0;
int dim = 0;
vector<vector<double>* > kernel_cache;
int multiclass_method = 1;

void my_assert(int assertion, char * msg){
    if(assertion)
	return;
    printf("Assertion : %s failed\n",msg);
    fflush(stdout);
    exit(15);
}


void exit_with_help()
{
    fprintf(stdout,
            "Usage: inc_multisvm [options] data_file result_file\n"
            "options:\n"
	    "-i initial training set size (default 1)\n"
	    "-s active selection type\n"
	    "   0 -- actively select from all three types of annotations\n"
	    "   1 -- randomly select from all three types of annotations\n"
	    "   2 -- actively select from instances (regions) alone\n"
	    "   3 -- randomly select from instances alone\n"
	    "   4 -- actively select from bags alone\n"
	    "   5 -- randomly select from bags alone\n"
	    "-P multiclass method (default 1)\n"
	    "-I active iterations\n"
	    "-L list of classes separated by commas (e.g: 1,2,4,6,8)\n"
	    "-T test split file\n"
	    "-V validation split file (for fitting a sigmoid)\n"
            "-B file format : files are stored in the following format:\n"
            "	0 -- libsvm ascii format (default)\n"
            "	1 -- binary format\n"
            "	2 -- split file format\n"
	    "   3 -- (-B 3 <training split file>)\n"
	    "-v 0|1 verbosity level (default 0)"
            "-o optimizer: set the type of optimization (default 1)\n"
            "	0 -- online \n"
            "	1 -- online with finishing step \n"
            "-t kernel_type : set type of kernel function (default 2)\n"
            "	0 -- linear: u'*v\n"
            "	1 -- polynomial: (gamma*u'*v + coef0)^degree\n"
            "	2 -- radial basis function: exp(-gamma*|u-v|^2)\n"
            "	3 -- sigmoid: tanh(gamma*u'*v + coef0)\n"
            "-s selection: set the type of selection strategy (default 0)\n"
            "	0 -- random \n" 
            "	1 -- gradient-based \n"
            "	2 -- margin-based \n"  
            "-T termination: set the type of early stopping strategy (default 0)\n"
            "	0 -- number of iterations \n" 
            "	1 -- number of SVs \n"
            "	2 -- time-based \n"  
            "-l sample: number of iterations/SVs/seconds to sample for early stopping (default all)\n"
            " if a list of numbers is given a model file is saved for each element of the set\n"
            "-C candidates : set number of candidates to search for selection strategy (default 50)\n" 
            "-d degree : set degree in kernel function (default 3)\n"
            "-g gamma : set gamma in kernel function (default 1/k)\n"
            "-r coef0 : set coef0 in kernel function (default 0)\n"
            "-c cost : set the parameter C of C-SVC\n"
            "-m cachesize : set cache memory size in MB (default 256)\n"
            "-wi weight: set the parameter C of class i to weight*C (default 1)\n"
            "-b bias: use a bias or not i.e. no constraint sum alpha_i y_i =0 (default 1=on)\n"
            "-e epsilon : set tolerance of termination criterion (default 0.001)\n"
            "-p epochs : number of epochs to train in online setting (default 1)\n"
            "-D deltamax : set tolerance for reprocess step, 1000=1 call to reprocess >1000=no calls to reprocess (default 1000)\n" 
        );
    exit(1);
}


void parse_command_line(int argc, char **argv, char *input_file_name, char *model_file_name)
{
    int i; int clss; double weight;

    // parse options
    for(i=1;i<argc;i++) 
    {
        if(argv[i][0] != '-') break;
        ++i;
        switch(argv[i-1][1])
        {
	case 'i':
	    initial_set_size = atoi(argv[i]);
	    break;
	case 'V':
	    strcpy(crossval_file_name,argv[i]);
	    break;
	case 'h':
	    threshold = atof(argv[i]);
	    break;
	case 'P':
	    multiclass_method = atoi(argv[i]);
	    break;
	case 'u':
	    strcpy(mi_un_split_file_,argv[i]);
	    break;
	case 'v':
	    verbose = atoi(argv[i]);
	    break;
	case 'L':
	    //strcpy(labels_file_name, argv[i]);
	    strcpy(class_labels,argv[i]);
	  break;
        case 'o':
            optimizer = atoi(argv[i]);
            break;
        case 't':
            kernel_type = atoi(argv[i]);
            break;
        case 's':
            mi_selection_type = atoi(argv[i]);
            break;
        case 'l':
            while(1)
            {
                select_size.push_back(atof(argv[i]));
                ++i; if((argv[i][0]<'0') || (argv[i][0]>'9')) break; 
            }
            i--;
            break;
	case 'd':
            degree = atof(argv[i]);
            break;
	case 'I':
	    active_iterations = atoi(argv[i]);
	    break;
        case 'g':
            kgamma = atof(argv[i]);
            break;
        case 'r':
            coef0 = atof(argv[i]);
            break;
	case 'R':
	    //r_tradeoff = atof(argv[i]);
	    r12 = atof(argv[i]);
	    //r21 = r12 * r_tradeoff;
	    break;
        case 'm':
            cache_size = (int) atof(argv[i]);
            break;
        case 'c':
            C = atof(argv[i]);
            break;
        case 'w':
            clss= atoi(&argv[i-1][2]);
            weight = atof(argv[i]);
            if (clss>=1) C_pos=weight; else C_neg=weight;
            break;
        case 'b':
            use_b0=atoi(argv[i]);
            break;
        case 'B':
            binary_files=atoi(argv[i]);
	    if (binary_files == 3)
		strcpy(mi_split_file_, argv[++i]);
	    break;
	case 'e':
            epsgr = atof(argv[i]);
            break;
        case 'p':
            epochs = atoi(argv[i]);
            break;
        case 'x':
            deltamax = atoi(argv[i]);
            break;
        case 'C':
            candidates = atoi(argv[i]);
            break;
        case 'S':
            termination_type = atoi(argv[i]);
            break;
	case 'F':
	    strcpy(model_file_name,argv[i]);
	    justtest = 1;
	    break;
	case 'D':
	    strcpy(fval_file_name, argv[i]);
	    break;
	case 'f':
	    strcpy(report_file_name, argv[i]);
	    break;
	case 'T':
	    strcpy(input_file_name, argv[i]);
	    strcpy(test_file_name, argv[i]);
	    break;
	default:
	    printf("UNKNOWN OPTION %c\n",argv[i-1][1]);
            fprintf(stderr,"unknown option\n");
            exit_with_help();
        }
    }

    saves=select_size.size(); 
    if(saves==0) select_size.push_back(100000000);

    // determine filenames

    if(i>=argc)
	if(justtest)
	    return;
	else
	    exit_with_help();
    
    strcpy(input_file_name, argv[i]);

    if(i<argc-1)
        strcpy(model_file_name,argv[i+1]);
    else
    {
        char *p = strrchr(argv[i],'/');
        if(p==NULL)
            p = argv[i];
        else
            ++p;
        sprintf(model_file_name,"%s.model",p);
    }

}





int split_file_load(char *f)
{
    int binary_file=0,labs=0,inds=0;
    FILE *fp;
    fp=fopen(f,"r"); 
    if(fp==NULL) {printf("[couldn't load split file: %s]\n",f); exit(1);}
    char dummy[200],dummy2[200];
    unsigned int i,j=0; for(i=0;i<strlen(f);i++) if(f[i]=='/') j=i+1;
    fscanf(fp,"%s %s",dummy,dummy2);
    strcpy(&(f[j]),dummy2);
    fscanf(fp,"%s %d",dummy,&binary_file);
    fscanf(fp,"%s %d",dummy,&inds);
    fscanf(fp,"%s %d",dummy,&labs);
    printf("[split file: load:%s binary:%d new_indices:%d new_labels:%d]\n",dummy2,binary_file,inds,labs);
    //printf("[split file:%s binary=%d]\n",dummy2,binary_file);
    if(!inds) return binary_file;
    while(1)
    {
        int i,j;
        int c=fscanf(fp,"%d",&i);
        if(labs) c=fscanf(fp,"%d",&j);
        if(c==-1) break;
        if (labs) 
            splits.push_back(ID(i-1,j)); 
        else 
            splits.push_back(ID(i-1,0));
    }

    sort(splits.begin(),splits.end());
	
    return binary_file;
}


int libsvm_load_data(char *filename)
// loads the same format as LIBSVM
{
    int index; double value;
    int elements, i;
    FILE *fp = fopen(filename,"r");
    lasvm_sparsevector_t* v;

    if(fp == NULL)
    {
        fprintf(stderr,"Can't open input file \"%s\"\n",filename);
        exit(1);
    }
    else
        printf("loading \"%s\"..  \n",filename);
    int splitpos=0;

    int msz = 0; 
    elements = 0;
    while(1)
    {
        int c = fgetc(fp);
        switch(c)
        {
        case '\n':
            if(splits.size()>0) 
            {
                if(splitpos<(int)splits.size() && splits[splitpos].x==msz)
                {
                    v=lasvm_sparsevector_create();
                    X.push_back(v);	splitpos++;
                }
            }
            else
            {
                v=lasvm_sparsevector_create();
                X.push_back(v); 
            }
            ++msz;
            //printf("%d\n",m);
            elements=0;
            break;
        case ':':
            ++elements;
            break;
        case EOF:
            goto out;
        default:
            ;
        }
    }
 out:
    rewind(fp);


    max_index = 0;splitpos=0;
    for(i=0;i<msz;i++)
    {

        int write=0;
        if(splits.size()>0)
        {
            if(splitpos<(int)splits.size() && splits[splitpos].x==i)
            {
                write=2;splitpos++;
            }
        }
        else
            write=1; 

        int label;
        fscanf(fp,"%d",&label);
        //	printf("%d %d\n",i,label);
        if(write) 
        {
            if(splits.size()>0)
            {  
                if(splits[splitpos-1].y!=0)
                    Y.push_back(splits[splitpos-1].y);
                else
                    Y.push_back(label);
            }
            else
                Y.push_back(label);
        }
			
        while(1)
        {
            int c;
            do {
                c = getc(fp);
                if(c=='\n') goto out2;
            } while(isspace(c));
            ungetc(c,fp);
            fscanf(fp,"%d:%lf",&index,&value);
			
            if (write==1) lasvm_sparsevector_set(X[m+i],index,value);
            if (write==2) lasvm_sparsevector_set(X[splitpos-1],index,value);
            if (index>max_index) max_index=index;
        }
    out2:
        label=1; // dummy
    }

    fclose(fp);

    msz=X.size()-m;
    printf("examples: %d   features: %d\n",msz,max_index);

    return msz;
}

int binary_load_data(char *filename)
{
    int msz,i=0,j;
    lasvm_sparsevector_t* v;
    int nonsparse=0;

    ifstream f;
    f.open(filename,ios::in|ios::binary);
    
    // read number of examples and number of features
    int sz[2]; 
    f.read((char*)sz,2*sizeof(int));
    if (!f) { printf("File writing error in line %d.\n",i); exit(1);}
    msz=sz[0]; max_index=sz[1];

    vector <float> val;
    vector <int>   ind;
    val.resize(max_index);
    if(max_index>0) nonsparse=1;
    int splitpos=0;

    for(i=0;i<msz;i++) 
    {
        int mwrite=0;
        if(splits.size()>0)
        {
            if(splitpos<(int)splits.size() && splits[splitpos].x==i) 
            { 
                mwrite=1;splitpos++;
                v=lasvm_sparsevector_create(); X.push_back(v);
            }
        }
        else
        { 
            mwrite=1;
            v=lasvm_sparsevector_create(); X.push_back(v);
        }
		
        if(nonsparse) // non-sparse binary file
        {
            f.read((char*)sz,1*sizeof(int)); // get label
            if(mwrite) 
            {
                if(splits.size()>0 && splits[splitpos-1].y!=0)
                    Y.push_back(splits[splitpos-1].y);
                else
                    Y.push_back(sz[0]);
            }
            f.read((char*)(&val[0]),max_index*sizeof(float));
            if(mwrite)
                for(j=0;j<max_index;j++) // set features for each example
                    lasvm_sparsevector_set(v,j,val[j]);
        }
        else			// sparse binary file
        {
            f.read((char*)sz,2*sizeof(int)); // get label & sparsity of example i
            if(mwrite) 
            {
                if(splits.size()>0 && splits[splitpos-1].y!=0)
                    Y.push_back(splits[splitpos-1].y);
                else
                    Y.push_back(sz[0]);
            }
            val.resize(sz[1]); ind.resize(sz[1]);
            f.read((char*)(&ind[0]),sz[1]*sizeof(int));
            f.read((char*)(&val[0]),sz[1]*sizeof(float));
            if(mwrite)
                for(j=0;j<sz[1];j++) // set features for each example
                {
                    lasvm_sparsevector_set(v,ind[j],val[j]);
                    //printf("%d=%g\n",ind[j],val[j]);
                    if(ind[j]>max_index) max_index=ind[j];
                }
        }		
    }
    f.close();

    msz=X.size()-m;
    printf("examples: %d   features: %d\n",msz,max_index);

    return msz;
}


void get_split_bags(const char* mi_split_file_,
		    map<int, MISparseVector*> &map_bags)
{
  ifstream stream_split(mi_split_file_);
  printf("split file is %s\n",mi_split_file_);
  assert(stream_split);

  string line;
  getline(stream_split, line);
  while (stream_split) {
    StripWhiteSpace(&line);
    map_bags[atoi(line.c_str())] = NULL;
    getline(stream_split, line);
  }
  stream_split.close();
}

int check_label(int label){
    int idx=-1;
    for(unsigned int i=0;i<classes.size();i++)
	if(label==classes[i]){
	    idx = i;
	    break;
	}
    if(idx!=-1)
	return 1;
    return 0;
}


int mi_load_data_train(const char* mi_file_name, const char* mi_split_file_)
{
  int msz = 0;
  
  // Read training bags ids from 'split_file'.
  map<int, MISparseVector*> map_bags;
  get_split_bags(mi_split_file_, map_bags);

  // Fill the bags with instances from the file.
  ifstream stream_in(mi_file_name);
  printf("mi file is %s\n",mi_file_name);
  assert(stream_in);

  // Ignore total number of bags.
  string line;
  getline(stream_in, line);
  assert(stream_in);

  // Read bag instances, one instance / line.
  getline(stream_in, line);
  while (stream_in) {
    string::size_type space = line.find(' ');
    assert(space != string::npos);
    string header = line.substr(0, space);
    string text = line.substr(space + 1);
    double acost = 50;
    double pcost = 50;
    char imgname[20] = "\0";

    // Process header.
    vector<string> header_tokens;
    SplitStringUsing(header, ":", &header_tokens);
    assert(header_tokens.size() == 3 || header_tokens.size() == 5 || header_tokens.size() == 4);


    int bag_id = atoi(header_tokens[1].c_str());
    int bag_label = atoi(header_tokens[2].c_str());

    if(header_tokens.size() == 5){
	pcost = atof(header_tokens[3].c_str());
	acost = atof(header_tokens[4].c_str());
    }

    if(header_tokens.size() == 4){
	strcpy(imgname, header_tokens[3].c_str());
    }
    

    //printf("bag %d Label %d Cost %lf\n",bag_id,bag_label,cost);
    if(!check_label(bag_label) && bag_label != 50){
	getline(stream_in, line);
	continue;
    }

    // Process only bags from the split
    map<int, MISparseVector*>::iterator it = map_bags.find(bag_id);
    if (it != map_bags.end()) {
	if (it->second == NULL) {
	  it->second = new MISparseVector;
	  it->second->label_ = bag_label;
	  it->second->bag_id = bag_id;
	  it->second->acost = acost;
	  it->second->pcost = pcost;
	  strcpy(it->second->imgname,imgname);
	}
      lasvm_sparsevector_t* instance = lasvm_sparsevector_create();
      instance->label = bag_label;
      instance->read(text);
      if(dim < instance->pairs[instance->size-1].index)
	  dim = instance->pairs[instance->size-1].index;
      it->second->addInstance(instance);
      if(instance->label==1)
	  it->second->label_ = 1;
    }
    
    getline(stream_in, line);
  }
  
  int numpos = 0,numneg = 0;
  // Create training examples from unpacked bags.
  map<int, MISparseVector*>::iterator it = map_bags.begin();
  while (it != map_bags.end()) {
    MISparseVector *bag = it->second;
    if(it->second == NULL){
	it++;
	continue;
    }
    if (bag->label_ == -100) {
	if(numneg <100000){
	    // Unpack negative bag in labeled singleton bags.
	    for (int i = 0; i < bag->size(); i++) {
		MISparseVector *bag_instance = new MISparseVector;
		bag_instance->addInstance(bag->getInstance(i));
		bag_instance->label_ = -1;
		X.push_back(bag_instance);
		Y.push_back(-1);
		msz++;
		data_map[TRAIN].push_back(X.size()-1);
		bag_pointer.push_back(NULL);
	    }
	    delete bag;
	    it->second = NULL;
	}
	else{
	    X.push_back(bag);
	    Y.push_back(-1);
	    msz++;
	    data_map[UNLABELEDBAGS].push_back(X.size()-1);
	    bag->index = X.size()-1;
	    bag_pointer.push_back(NULL);
	}
	numneg++;
    }
    else{
	X.push_back(bag);
	Y.push_back(+1);
	msz++;
	data_map[UNLABELEDBAGS].push_back(X.size()-1);
	numpos++;
	bag->index = X.size()-1;
	bag_pointer.push_back(NULL);
	got_label.push_back(0);
    }
    it++;
  }


  for(unsigned int i = 0; i < data_map[UNLABELEDBAGS].size(); i++) {
      MISparseVector *misv_i;
      misv_i = (MISparseVector *) X[data_map[UNLABELEDBAGS][i]];
      for (int k = 0; k < misv_i->size(); k++) {
	  MISparseVector *bag_instance = (MISparseVector *)misv_i->getInstance(k);
	  int found = 0;
	  for(unsigned int l =0;l<misv_i->labels.size();l++)
	      if(bag_instance->label == misv_i->labels[l]){
		  found = 1;
		  break;
	      }
	  if(!found && check_label(bag_instance->label)){// || bag_instance->label==50){
	      misv_i->labels.push_back(bag_instance->label);
	      misv_i->acquired.push_back(0);
	  }
      }
  }


  printf("Examples loaded %d: pos:%d neg: %d\n",msz,numpos,numneg);
  return msz;
}


int mi_load_data_unlabeled(const char* mi_file_name, const char* mi_un_split_file_)
{
  int msz = 0;
  
  if(strcmp(mi_un_split_file_,"none")==0)
      return msz;
  // Read training bags ids from 'split_file'.
  map<int, MISparseVector*> map_bags;
  get_split_bags(mi_un_split_file_, map_bags);

  // Fill the bags with instances from the file.
  ifstream stream_in(mi_file_name);
  printf("mi file is %s\n",mi_file_name);
  assert(stream_in);

  // Ignore total number of bags.
  string line;
  getline(stream_in, line);
  assert(stream_in);

  // Read bag instances, one instance / line.
  getline(stream_in, line);
  while (stream_in) {
    string::size_type space = line.find(' ');
    assert(space != string::npos);
    string header = line.substr(0, space);
    string text = line.substr(space + 1);
    
    // Process header.
    vector<string> header_tokens;
    SplitStringUsing(header, ":", &header_tokens);
    assert(header_tokens.size() == 3);

    int bag_id = atoi(header_tokens[1].c_str());
    int bag_label = atoi(header_tokens[2].c_str());

    // Process only bags from the split
    map<int, MISparseVector*>::iterator it = map_bags.find(bag_id);
    if (it != map_bags.end()) {
      if (it->second == NULL) {
	it->second = new MISparseVector;
	it->second->label_ = bag_label;
	it->second->bag_id = bag_id;
      }

      lasvm_sparsevector_t* instance = lasvm_sparsevector_create();
      instance->label = bag_label;
      instance->read(text);
      if(dim < instance->pairs[instance->size-1].index)
	  dim = instance->pairs[instance->size-1].index;
      it->second->addInstance(instance);
      if(instance->label==1)
	  it->second->label_ = 1;
    }
    
    getline(stream_in, line);
  }

  // Create training examples from unpacked bags.
  map<int, MISparseVector*>::iterator it = map_bags.begin();
  while (it != map_bags.end()) {
    MISparseVector *bag = it->second;
    if (bag->label_ == -1) {
      X.push_back(bag);
      Y.push_back(-1);
      msz++;
      data_map[UNLABELEDBAGS].push_back(X.size()-1);
      bag->index = X.size()-1;
      bag_pointer.push_back(NULL);
    }
    else {
      // Record positive bag.
      X.push_back(bag);
      Y.push_back(+1);
      msz++;
      data_map[UNLABELEDBAGS].push_back(X.size()-1);
      bag->index = X.size()-1;
      bag_pointer.push_back(NULL);
    }
    it++;
  }

  printf("Examples loaded %d\n",msz);
  return msz;
}



int mi_load_data_test(const char* mi_file_name, const char* mi_split_file_)
{
  int msz = 0;
  
  // Read training bags ids from 'split_file'.
  map<int, MISparseVector*> map_bags;
  get_split_bags(mi_split_file_, map_bags);

  // Fill the bags with instances from the file.
  ifstream stream_in(mi_file_name);
  assert(stream_in);

  // Ignore total number of bags.
  string line;
  getline(stream_in, line);
  assert(stream_in);

  // Read bag instances, one instance / line.
  getline(stream_in, line);
  while (stream_in) {
    string::size_type space = line.find(' ');
    assert(space != string::npos);
    string header = line.substr(0, space);
    string text = line.substr(space + 1);
    
    // Process header.
    vector<string> header_tokens;
    SplitStringUsing(header, ":", &header_tokens);
    assert(header_tokens.size() == 3);
    // Ignore instance ID
    atoi(header_tokens[0].c_str());
    int bag_id = atoi(header_tokens[1].c_str());
    int bag_label = atoi(header_tokens[2].c_str());
    if(!check_label(bag_label)){
	getline(stream_in, line);
	continue;
    }
    // Process only bags from the split
    map<int, MISparseVector*>::iterator it = map_bags.find(bag_id);
    if (it != map_bags.end()) {
      if (it->second == NULL) {
	it->second = new MISparseVector;
	it->second->label_ = bag_label;
      }

      lasvm_sparsevector_t* instance = lasvm_sparsevector_create();
      StripWhiteSpace(&text);
      instance->read(text);
      instance->label = bag_label;
      if(dim < instance->pairs[instance->size-1].index)
	  dim = instance->pairs[instance->size-1].index;
      it->second->addInstance(instance);
      if(instance->label==1)
	  it->second->label_ = 1;
    }
    
    getline(stream_in, line);
  }
  
  // Create training examples from unpacked bags.
  map<int, MISparseVector*>::iterator it = map_bags.begin();
  while (it != map_bags.end()) {
      if(it->second ==NULL){
	  printf("it->second is NULL\n");
	  it++;
	  continue;
      }
    X.push_back(it->second);
    Y.push_back(it->second->label_);
    bag_pointer.push_back(NULL);
    msz++;
    data_map[TEST].push_back(X.size()-1);
    it++;
  }

  for (unsigned int i = 0; i < data_map[TEST].size(); i++) {
      MISparseVector *misv_i = (MISparseVector*)(X[data_map[TEST][i]]);
      for (int k = 0; k < misv_i->size(); k++) {
	  MISparseVector *bag_instance = (MISparseVector *)misv_i->getInstance(k);
	  int found = 0;
	  for(unsigned int l =0;l<misv_i->labels.size();l++)
	      if(bag_instance->label == misv_i->labels[l]){
		  found = 1;
		  break;
	      }
	  if(!found){
	      misv_i->labels.push_back(bag_instance->label);
	      misv_i->acquired.push_back(0);
	  }
      }
  }
  
  return msz;
}



int mi_load_data_crossval(const char* mi_file_name, const char* mi_split_file_)
{
  int msz = 0;
  // Read training bags ids from 'split_file'.
  map<int, MISparseVector*> map_bags;
  get_split_bags(mi_split_file_, map_bags);

  // Fill the bags with instances from the file.
  ifstream stream_in(mi_file_name);
  assert(stream_in);

  // Ignore total number of bags.
  string line;
  getline(stream_in, line);
  assert(stream_in);

  

  // Read bag instances, one instance / line.
  getline(stream_in, line);
  while (stream_in) {
    string::size_type space = line.find(' ');
    assert(space != string::npos);
    string header = line.substr(0, space);
    string text = line.substr(space + 1);
    
    // Process header.
    vector<string> header_tokens;
    SplitStringUsing(header, ":", &header_tokens);
    assert(header_tokens.size() == 3);
    // Ignore instance ID
    atoi(header_tokens[0].c_str());
    int bag_id = atoi(header_tokens[1].c_str());
    int bag_label = atoi(header_tokens[2].c_str());
    if(!check_label(bag_label)){
	getline(stream_in, line);
	continue;
    }
    // Process only bags from the split
    map<int, MISparseVector*>::iterator it = map_bags.find(bag_id);
    if (it != map_bags.end()) {
      if (it->second == NULL) {
	it->second = new MISparseVector;
	it->second->label_ = bag_label;
      }

      lasvm_sparsevector_t* instance = lasvm_sparsevector_create();
      StripWhiteSpace(&text);
      instance->read(text);
      instance->label = bag_label;
      if(dim < instance->pairs[instance->size-1].index)
	  dim = instance->pairs[instance->size-1].index;
      it->second->addInstance(instance);
      if(instance->label==1)
	  it->second->label_ = 1;
    }
    
    getline(stream_in, line);
  }

  // Create training examples from unpacked bags.
  map<int, MISparseVector*>::iterator it = map_bags.begin();
  while (it != map_bags.end()) {
      if(it->second ==NULL){
	  printf("it->second is NULL\n");
	  it++;
	  continue;
      }
      MISparseVector * misv_i = it->second;
      for(unsigned int k=0;k<misv_i->size();k++){
	  MISparseVector *bag_instance = (MISparseVector *)misv_i->getInstance(k);
	  MISparseVector *bag = new MISparseVector;
	  bag->addInstance(bag_instance);
	  bag->label_ = bag_instance->label;
	  bag->bag_id = misv_i->bag_id;
	  X.push_back(bag);
	  Y.push_back(bag_instance->label);
	  msz++;
	  bag->index = X.size()-1;
	  bag_instance->index = X.size()-1;
	  bag_pointer.push_back(misv_i);
	  data_map[CROSSVAL].push_back(X.size()-1);
      }
    it++;
  }
  
  return msz;
}


void load_positive_bag_labels(){
    if(labels_file_name[0] == '\0')
	return;
    int flag = 0;
    ifstream stream_in(labels_file_name);
    if(strcmp(labels_file_name,"none")!=0){
	printf("labels file is %s\n",labels_file_name);
	assert(stream_in);
	flag = 1;
    }

  string line;

  int count = 0;
  for (unsigned int i = 0; i < data_map[TRAINPOS].size(); i++) {
      MISparseVector *misv_i;
      misv_i = (MISparseVector *) X[data_map[TRAINPOS][i]];
      for (int k = 0; k < misv_i->size(); k++) {
	MISparseVector *bag_instance = (MISparseVector *)misv_i->getInstance(k);

	if(flag){
	    getline(stream_in, line);
	    bag_instance->label =  atoi(line.c_str());
	}
	MISparseVector *bag = new MISparseVector;
	bag->addInstance(bag_instance);
	bag->label_ = bag_instance->label;
	bag->bag_id = misv_i->bag_id;
	bag->acost = misv_i->acost/misv_i->size();
	bag->pcost = misv_i->pcost/misv_i->size();
	X.push_back(bag);
	Y.push_back(bag_instance->label);
	m++;
	count++;
	bag->index = X.size()-1;
	bag_instance->index = X.size()-1;
	misv_i->indices.push_back(X.size()-1);
	if(check_label(bag_instance->label) || bag_instance->label == 50)
	    data_map[POSINSTANCES].push_back(X.size()-1);
	bag_pointer.push_back(misv_i);
      }
  }    

  for (unsigned int i = 0; i < data_map[UNLABELEDBAGS].size(); i++) {
      MISparseVector *misv_i;
      misv_i = (MISparseVector *) X[data_map[UNLABELEDBAGS][i]];
      for (int k = 0; k < misv_i->size(); k++) {
	  MISparseVector *bag_instance = (MISparseVector *)misv_i->getInstance(k);
	  if(flag){
	      getline(stream_in, line);
	      bag_instance->label =  atoi(line.c_str());
	  }
	  MISparseVector *bag = new MISparseVector;
	  bag->addInstance(bag_instance);
	  bag->label_ = bag_instance->label;
	  bag->bag_id = misv_i->bag_id;
	  bag->acost = misv_i->acost/misv_i->size();
	  bag->pcost = misv_i->pcost/misv_i->size();
	  X.push_back(bag);
	  Y.push_back(bag_instance->label);
	  m++;
	  count++;
	  bag->index = X.size()-1;
	  bag_instance->index = X.size()-1;
	  misv_i->indices.push_back(X.size()-1);
	  bag_pointer.push_back(misv_i);
      }
  }    
  
  if(flag)
      stream_in.close();
}


void load_tmp_data(){
    int numpos = 0,numneg=0;
    int set = TEST;
    for (unsigned int i = 0; i < data_map[set].size(); i++) {
      MISparseVector *misv_i;
      misv_i = (MISparseVector *) X[data_map[set][i]];
      for (int k = 0; k < misv_i->size(); k++) {
	  MISparseVector *bag_instance = (MISparseVector *)misv_i->getInstance(k);
	  MISparseVector *bag = new MISparseVector;
	  bag->addInstance(bag_instance);
	  bag->label_ = bag_instance->label;
	  X.push_back(bag);
	  bag_pointer.push_back(NULL);
	  Y.push_back(bag_instance->label);
	  misv_i->indices.push_back(X.size()-1);
	  m++;
	  data_map[TMP].push_back(X.size()-1);
	  if(bag_instance->label==1)
	      numpos++;
	  else
	      numneg++;
      }
    }
    //printf("NUMPOS %d NUMNEG %d\n",numpos,numneg);
}


void load_data_file(char *filename)
{
    int msz,i,ft;
    splits.resize(0); 

    int bin=binary_files;
    if(bin==0) // if ascii, check if it isn't a split file..
    {
        FILE *f=fopen(filename,"r");
        if(f == NULL)
        {
            fprintf(stderr,"Can't open input file \"%s\"\n",filename);
            exit(1);
        }
        char c; fscanf(f,"%c",&c); 
        if(c=='f') bin=2; // found split file!
    }

    switch(bin)  // load diferent file formats
    {
    case 0: // libsvm format
        msz=libsvm_load_data(filename); break;
    case 1: 
        msz=binary_load_data(filename); break;
    case 2:
        ft=split_file_load(filename);
        if(ft==0) 	 
        {msz=libsvm_load_data(filename); break;} 
        else
        {msz=binary_load_data(filename); break;}
    case 3:
	if(!justtest)
	    msz = mi_load_data_train(filename, mi_split_file_);
	else
	    msz = mi_load_data_test(filename, mi_split_file_);
	break;
    default:
        fprintf(stderr,"Illegal file type '-B %d'\n",bin);
        exit(1);
    }

    if(kernel_type==RBF)
    {
        x_square.resize(m+msz);
        for(i=0;i<msz;i++)
            x_square[i+m]=lasvm_sparsevector_dot_product(X[i+m],X[i+m]);
    }

    if(kgamma==-1)
        kgamma=1.0/ ((double) max_index); // same default as LIBSVM

    m+=msz;
    load_positive_bag_labels();
    if(strcmp(test_file_name,"none")!=0)
	m+=mi_load_data_test(filename, test_file_name);
    mi_load_data_crossval(filename, crossval_file_name);
    m+=data_map[CROSSVAL].size();
    load_tmp_data();
}


int load_model(const char *model_file_name)
{
  ifstream stream_in(model_file_name);
  if (!stream_in.good()) {
    fprintf(stderr,"Model file does not exist!\n");
    exit(1);
  }
  
  double alpha0 = 0.0;
  int num_sv = 0;

  // Read parameters.
  string line;
  getline(stream_in, line);
  while (stream_in) {
    vector<string> line_tokens;
    SplitStringUsing(line, " ", &line_tokens);
    string cmd = line_tokens[0];
    if ( cmd == "b0") {
      b0 = atof(line_tokens[1].c_str());
      printf("\tb0: %g\n",b0);
    }
    else if (cmd == "kernel_type") {
      string k_type = line_tokens[1];
      int i;
      for (i = 0; kernel_type_table[i]; ++i) {
	if (k_type == kernel_type_table[i]) {
	  kernel_type=i;
	  break;
	}
      }
      if (i == sizeof(kernel_type_table)) {
	fprintf(stderr,"kernel_type not know\n");
	exit(1);
      }
      else
	printf("\tkernel_type: %s\n", kernel_type_table[kernel_type]);
    }
    else if (cmd == "degree") {
      degree = atof(line_tokens[1].c_str());
      printf("\tdegree: %g\n", degree);
    }
    else if (cmd == "kgamma") {
      kgamma = atof(line_tokens[1].c_str());
      printf("\tkgamma: %g\n", kgamma);
    }
    else if (cmd == "coef0") {
      coef0 = atof(line_tokens[1].c_str());
      printf("\tcoef0: %g\n", coef0);
    }
    else if (cmd == "alpha0") {
      alpha0 = atof(line_tokens[1].c_str());
      printf("\talpha0: %g\n", alpha0);
    }
    else if (cmd == "total_sv") {
      num_sv = atoi(line_tokens[1].c_str());
      printf("\tNumber of SVs: %d\n", num_sv);
    }
    else if (cmd == "SV") {
      break;
    }
    else {
      fprintf(stderr, "unknown text in model file\n");
      exit(1);
    }

    getline(stream_in, line);
  }

  // Read sv_coef and SV
  int msz = 0;
  // Get bag line.
  getline(stream_in, line);
  while (stream_in) {
    // Read alpha.
    string::size_type space = line.find(' ');
    assert(space != string::npos);
    double tmp_alpha = atof(line.substr(0, space).c_str());
    alpha.push_back(tmp_alpha);

    // Read bag.
    string header = line.substr(space + 1);
    StripWhiteSpace(&header);
    int bag_size = atoi(header.c_str());
    MISparseVector *bag = new MISparseVector();
    // Read instances.
    for (int i = 0; i < bag_size; i++) {
      getline(stream_in, line);
      lasvm_sparsevector_t* instance = lasvm_sparsevector_create();
      StripWhiteSpace(&line);
      instance->read(line);
      bag->addInstance(instance);
    }
    X.push_back(bag);

    if (tmp_alpha != 0) 
      data_map[SV].push_back(m + msz);
    else
      data_map[UNLABELED].push_back(m + msz);

    msz++;

    // Get next bag line.
    getline(stream_in, line);
  }
    
  stream_in.close();
  
  if (data_map[UNLABELED].size() > 0)
    alpha.push_back(alpha0);

  assert (msz == ((int) X.size()) - m);
  printf("examples: %d\n", msz);
  printf("|SV|: %d   |unlabeled|: %d\n",
	 data_map[SV].size(),
	 data_map[UNLABELED].size());
  return msz;
}



int sv1,sv2; double max_alpha,alpha_tol;

int count_svs()
{
    int i; 
    max_alpha=0; 
    sv1=0;sv2=0;
    
    for(i=0;i<m;i++) 	// Count svs..   
    {
        if(alpha[i]>max_alpha) max_alpha=alpha[i];
        if(-alpha[i]>max_alpha) max_alpha=-alpha[i];
    }
       
    alpha_tol=max_alpha/1000.0;
    
    for(i=0;i<m;i++) 
    {
        if(Y[i]>0) 
        {
            if(alpha[i] >= alpha_tol) sv1++; 
        }
        else    
        {
            if(-alpha[i] >= alpha_tol) sv2++; 
        }            
    }
    return sv1+sv2;
}


void print_model_file(char* model_file_name)
{
  ofstream stream_out(model_file_name);
  assert(stream_out.good());

  stream_out << "b0 " << b0 << endl;
  stream_out << "kernel_type " << kernel_type_table[kernel_type] << endl;
  stream_out << "kgamma " << kgamma << endl;
  stream_out << "degree " << degree << endl;
  stream_out << "coef0 " << coef0 << endl;

  // write out balancing constraint alpha
  if ((int) alpha.size() == m + 1)
    stream_out << "alpha0 " << alpha[m] << endl;

  int num_sv = 0;
  for (int i = 0; i < m; ++i) {
    if (alpha[i] != 0)
      ++num_sv;
  }
  stream_out << "total_sv " << num_sv << endl;
  stream_out << "SV" << endl;

  // write out SVs
  for (int i = 0; i < m; ++i) {
    if (alpha[i] != 0) {
      stream_out << alpha[i] << " ";
      X[i]->write(stream_out);
    }
  }
  
  stream_out.close();
}

double kernel(MISparseVector *misv1, MISparseVector *misv2,int i, int j)
{
  double dot = misv1->dot_product(misv2) / (misv1->norm() * misv2->norm());
  double x1_square=0;
  double x2_square=0;
  // sparse, linear kernel
  switch (kernel_type) {
    case LINEAR:
      break;
    case POLY:
      dot = pow(kgamma*dot+coef0,degree);
      break;
    case RBF:
	x2_square = misv2->dot_product(misv2) / (misv2->norm() * misv2->norm());
	x1_square = misv1->dot_product(misv1) / (misv1->norm() * misv1->norm());
	dot = exp(-kgamma*(x1_square+x2_square-2*dot));
	break;
    case SIGMOID:
      dot = tanh(kgamma*dot+coef0);
      break;
  }
  kcalcs++;
  return dot;
}


double saveStatisticsROC(LabelConfidence* lc, int size, bool write)
{
  double ot=0;
  int flag = 1;
  // Sort confidence vector, output counts for all confidence levels.
  sort(lc, lc + size);

  // Compute total true positives
  double total_pos = 0;
  for (int i = 0; i < size; i++) {
    if (lc[i].label_ == 1)
      total_pos++;
  }
  double total_neg = size - total_pos;

  int true_pos = 0;
  int false_pos = 0;
  double area = 0.0;
  double fp_rate = 0.0;
  double tp_rate = 0.0;

  ofstream stream_out;
  if(write){
      stream_out.open(report_file_name);
      assert(stream_out);
  }
  for (int i = 0; i < size; i++) {
    if (lc[i].label_ == 1) {
      true_pos++;
      tp_rate = true_pos / total_pos;
    }
    else {
      false_pos++;
      area += tp_rate / total_neg;
      fp_rate = false_pos / total_neg;
    }
    if(flag==1 && tp_rate+fp_rate >= 1){
      ot = lc[i].conf_;
      flag = 0;
    }
    if(write)
	stream_out << fp_rate << " " << tp_rate << endl;
  }
  if(write){
      stream_out << "area = " << area << endl;
      stream_out << "ot = " << ot << endl;
      stream_out.close();
  }
  return area;
}



double saveStatisticsMRR(LabelConfidence* lc, int size, bool write){
    double total_pos = 0;
    double threshold = saveStatisticsROC(lc,size,0);
    for (int i = 0; i < size; i++) {
	if (lc[i].label_ == 1)
	    total_pos++;
    }
    double total_neg = size - total_pos;
    double correct_pos = 0;
    double correct_neg = 0;
    for(int i=0;i<size;i++){
	if(lc[i].conf_ >= threshold && lc[i].label_==1)
	    correct_pos++;
	if(lc[i].conf_ < threshold && lc[i].label_==-1)
	    correct_neg++;
    }
    double mrr = (correct_pos/total_pos + correct_neg/total_neg)*50;
    ofstream stream_out;
    if(write){
	stream_out.open(report_file_name);
	assert(stream_out);
	stream_out << mrr << " "<<threshold<<endl;
	stream_out.close();
    }
    cout << "mrr = "<<mrr<<endl;
    
    return mrr;
}


void saveStatisticsPR(LabelConfidence* lc, int size)
{
  // Sort confidence vector, output counts for all confidence levels.
  sort(lc, lc + size);

  // Compute total true positives
  int total_true = 0;
  for (int i = 0; i < size; i++) {
    if (lc[i].label_ == 1)
      total_true++;
  }

  int true_true = 0;
  int false_true = 0;

  ofstream stream_out(report_file_name);
  assert(stream_out);
  stream_out << size << " " << total_true << endl;
  for (int i = 0; i < size; i++) {
    if (lc[i].label_ == 1)
      true_true++;
    else
      false_true++;
    stream_out << lc[i].conf_ << " " << true_true << " " << false_true << endl;
  }
  stream_out.close();
}

double predict_value(IncSVM* svm, int idx);


double test_model(IncSVM* svm, int set)
{
    printf("Testing model ...\n");
    bool write = false;

    // don't add the ridge while testing on training data
    ofstream stream_out;
    if(write){
	stream_out.open(fval_file_name);
	assert(stream_out);
    }
    int cf[classes.size()][classes.size()];
    for(unsigned int i=0;i<classes.size();i++)
	for(unsigned int j=0;j<classes.size();j++)
	    cf[i][j] = 0;

    for(unsigned int i=0;i<data_map[set].size();i++){
	int label = svm->classify(data_map[set][i]);
	stream_out<<label<<endl;
	cf[rev_idx[Y[data_map[set][i]]]][rev_idx[label]]++;
    }

    stream_out.close();
    double acc = 0;
    double acc1 = 0;
    double total = 0;
    for(unsigned int i=0;i<classes.size();i++){
	double sum = 0;
	for(unsigned int j=0;j<classes.size();j++){
	    sum += cf[i][j];
	    printf("%5d",cf[i][j]);
	}
	acc += (double)cf[i][i]/sum;
	acc1 += cf[i][i];
	total += sum;
	printf("\n");
    }
    cout<<"Per-class mean accuracy = "<<acc/classes.size()*100<<endl;
    cout<<"Global mean accuracy = "<<(double)acc1/total*100<<endl;
    return (double)acc1/total*100;
}


double calculate_risk(IncSVM* svm, int set)
{
    printf("Calculating risk ...\n");
    double risk = 0;
    double probs[classes.size()];

    for(unsigned int i=0;i<data_map[set].size();i++){
	svm->svmprobs(data_map[set][i],probs);
	risk += 1 - probs[rev_idx[Y[data_map[set][i]]]];
    }
    cout<<"Risk = "<<risk<<endl;
    return risk;
}


double test_current_model(IncSVM* svm, int type,int set)
{
    bool write = false;
    if(set == TEST || set == TMP)
	write = true;
  // don't add the ridge while testing on training data
  LabelConfidence *lc = new LabelConfidence[data_map[set].size()];
  ofstream stream_out;
  if(write){
      stream_out.open(fval_file_name);
      assert(stream_out);
  }

  double alp[m];
  double bx = b0;
  double sum = 0;
  if(type){
      bx = svm->b;
      for(int j=0;j<m;j++)
	  alp[j] = 0;
      
      int sets[] = {0,1,2,3};
      for(int i=0;i<3;i++)
	  for(unsigned int j=0;j<svm->ind[sets[i]].size();j++)
	      if(sets[i]<2 || svm->a[svm->ind[sets[i]][j]]>0)
		  alp[svm->ind[sets[i]][j]] = Y[svm->ind[sets[i]][j]] * svm->a[svm->ind[sets[i]][j]];
  }
  else{
      for(int j=0;j<m;j++){
	  alp[j] = alpha[j];
	  sum += alp[j];
      }
  }
  
  for (unsigned int i = 0; i < data_map[set].size(); i++) {
    MISparseVector *misv_i = (MISparseVector *) X[data_map[set][i]];
    double max_score = -std::numeric_limits<double>::infinity();
    double score = 0;
    for (int k = 0; k < misv_i->size(); k++) {
      MISparseVector *bag_instance = new MISparseVector();
      bag_instance->addInstance(misv_i->getInstance(k));
      
      double y = bx;
      double a_max = -1;
      
      // calc remaining part of decision function
      for (int j = 0; j < m; j++) {
	  if (alp[j] == 0)
	      continue;
	  double kval = kernel(bag_instance, (MISparseVector *) X[j],-1,j);
	  y += alp[j] * kval;
	  if(alp[j]>a_max)
	      a_max = alp[j];
      }
      if(i==0 && k==0)
	  printf("Max is %lf\n",a_max);
      
      // Update max score.
      if (y > max_score)
	max_score = y;
      // Update bag score.
      if (y > -1)
	score += y + 1;

      delete bag_instance;
    }
    
    if(write)
	stream_out << max_score << endl;
    lc[i].label_ = Y[data_map[set][i]];
    lc[i].conf_ = max_score;
  }
  if(write)
      stream_out.close();
  // Save statistics needed for PR curves
  double area = saveStatisticsROC(lc, data_map[set].size(),write);
  delete[] lc;
  // set back ker_ridge
  return area;
}




double test_label_presence(IncSVM* svm, int label,int set)
{
    LabelConfidence *lc = new LabelConfidence[data_map[set].size()];
    for (unsigned int i = 0; i < data_map[set].size(); i++) {
	MISparseVector * misv_i = (MISparseVector *)(X[data_map[set][i]]);
	int present = -1;
	for (unsigned int k = 0; k < misv_i->labels.size(); k++) 
	    if(misv_i->labels[k] == label){
		present = 1;
		break;
	    }
	lc[i].label_ = present;
	lc[i].conf_ = svm->svmprob_ispresent(&misv_i->indices,label);
    }
    double area = saveStatisticsROC(lc, data_map[set].size(),false);
    delete[] lc;
    return area;
}


double test_label_presence1(IncSVM* svm, int set,double* area)
{
    double avg = 0;
    double probs[classes.size()];
    LabelConfidence **lc = new LabelConfidence*[classes.size()];
    for(unsigned int i=0;i<classes.size();i++)
	lc[i] = new LabelConfidence[data_map[set].size()];

    for (unsigned int i = 0; i < data_map[set].size(); i++) {
	MISparseVector * misv_i = (MISparseVector *)(X[data_map[set][i]]);
	for(unsigned int j=0;j<classes.size();j++)
	    lc[j][i].label_ = -1;

	for (unsigned int k = 0; k < misv_i->labels.size(); k++) {
	    int idx = rev_idx[misv_i->labels[k]];
	    lc[idx][i].label_ = 1;
	}

	svm->svmprobs_ispresent(&misv_i->indices,probs);
	for(unsigned int j=0;j<classes.size();j++){
	    lc[j][i].conf_ = probs[j];
	}
    }

    for(unsigned int i=0;i<classes.size();i++){
	area[i] = saveStatisticsROC(lc[i], data_map[set].size(),false);
	avg = avg + area[i];
	delete[] lc[i];
    }
    delete lc;
    return avg/classes.size();
}



double test_region_accuracy(IncSVM* svm, int set,double* area)
{
    double avg = 0;
    double probs[classes.size()];
    LabelConfidence **lc = new LabelConfidence*[classes.size()];
    for(unsigned int i=0;i<classes.size();i++)
	lc[i] = new LabelConfidence[data_map[set].size()];

    for (unsigned int i = 0; i < data_map[set].size(); i++) {
	MISparseVector * misv_i = (MISparseVector *)(X[data_map[set][i]]);
	for(unsigned int j=0;j<classes.size();j++)
	    lc[j][i].label_ = -1;
	
	int idx = rev_idx[Y[data_map[set][i]]];
	lc[idx][i].label_ = 1;

	svm->svmprobs(data_map[set][i],probs);
	for(unsigned int j=0;j<classes.size();j++)
	    lc[j][i].conf_ = probs[j];
    }

    for(unsigned int i=0;i<classes.size();i++){
	area[i] = saveStatisticsROC(lc[i], data_map[set].size(),false);
	avg = avg + area[i];
	delete[] lc[i];
    }
    delete lc;
    return avg/classes.size();
}


double average_auroc(IncSVM* svm, int set){
    double avg_auroc = 0;
    for(unsigned int i=0;i<classes.size();i++)
	avg_auroc += test_label_presence(svm, classes[i], set);
    avg_auroc /= classes.size();
    return avg_auroc;
}


void reset_kernel_row(int i){
    if(i<(int)kernel_cache.size())
	for(unsigned int j=0;j<kernel_cache.size();j++){
	    (*kernel_cache[i])[j] = -1;
	    (*kernel_cache[j])[i] = -1;
	}
}

void add_kernel_row(){
    vector<double> * tmp = new vector<double>;
    for(unsigned int j=0;j<kernel_cache.size();j++){
	(*kernel_cache[j]).push_back(-1);
	tmp->push_back(-1);
    }
    tmp->push_back(-1);
    kernel_cache.push_back(tmp);
}




double kernel(int i, int j, void *kparam)
{
    double dot,x1_square,x2_square;
    kcalcs++;
    if(i<(int)kernel_cache.size() && j<(int)(*kernel_cache[i]).size() && (*kernel_cache[i])[j] != -1)
      return (*kernel_cache[i])[j];
    MISparseVector *misv1 = (MISparseVector *) X[i];
    MISparseVector *misv2 = (MISparseVector *) X[j];
    dot = misv1->dot_product(misv2) / (misv1->norm() * misv2->norm());

    // sparse, linear kernel
    switch(kernel_type)
    {
    case LINEAR:
	break;
    case POLY:
	dot=pow(kgamma*dot+coef0,degree);
	break;
    case RBF:
	x1_square = misv1->dot_product(misv1) / (misv1->norm() * misv1->norm());
	x2_square = misv2->dot_product(misv2) / (misv2->norm() * misv2->norm());
	dot = exp(-kgamma*(x1_square+x2_square-2*dot));
	break;
    case SIGMOID:
        dot = tanh(kgamma*dot+coef0);    
	break;
    }
    
    if(i<(int)kernel_cache.size() && j<(int)(*kernel_cache[i]).size()){
	(*kernel_cache[i])[j] = dot;
	(*kernel_cache[j])[i] = dot;
    }
    return dot;
} 
  


/******************************* Incremental svm interface methods *********************************************/

void load_data_incremental(IncSVM* svm){
    for(unsigned int i=0;i<X.size();i++){
	svm->adddata(i,Y[i],C*X[i]->norm());
    }
}

int check(int,int);

void process_instance(IncSVM* svm, int idx, int label){
    //printf("Processing instance %d with label %d\n",idx,label);
    svm->setY(idx,label);
    stopwatch sw;
    ((IncSVM_posterior*)svm)->learn(idx,1);
    //printf("Time to learn %lf\n",sw.get_time());
}

void unlearn_instance(IncSVM* svm, int idx, int label){
    //printf("Unlearning instance %d with label %d\n",idx,label);
    stopwatch * sw = new stopwatch;
    ((IncSVM_posterior*)svm)->unlearn(idx);
    //printf("Time to unlearn %lf\n",sw->get_time());
    delete sw;
    svm->setY(idx,label);
}

void train_incremental(IncSVM* svm){
    int set = TRAIN;
    for(unsigned int i=0;i<data_map[set].size()-1;i++){
	//printf("Processing instance %d with label %d\n",data_map[set][i],Y[data_map[set][i]]);
	process_instance(svm,data_map[set][i],Y[data_map[set][i]]);
	//printf("Done\n");
    }
}

double predict_value(IncSVM* svm, int idx){
    return svm->svmeval(idx,NULL);
}

/******************************** Active Selection Routines *********************************************/


void evaluateSet(IncSVM* svm, int Set, double * out,int *labels,double* numpos, double* numneg){
    stopwatch sw;
    for(unsigned int i=0;i<data_map[Set].size();i++){
	out[i] = svm->svmeval(data_map[Set][i],NULL);
	if(labels!=NULL){
	    labels[i] = Y[data_map[Set][i]];
	    if(labels[i]==1)
		(*numpos)++;
	    else if(labels[i]==-1)
		(*numneg) = (*numneg)+1;
	}
    }
}


void evaluateSetUSVM(int Set, double * out,int *labels,double* numpos, double* numneg){
    for(unsigned int i=0;i<data_map[Set].size();i++){
	double y = b0;
	
	// calc remaining part of decision function
	for (int j = 0; j < m; j++) {
	    if (alpha[j] == 0)
		continue;
	    y += alpha[j] * kernel((MISparseVector*)X[data_map[Set][i]], (MISparseVector *) X[j],-1,j);
	}

	out[i] = y;
    if(labels!=NULL){
      labels[i] = Y[data_map[Set][i]];
      if(labels[i]==1)
	(*numpos)++;
      else if(labels[i]==-1)
	(*numneg)++;
    }
  }
}

double min(double a, double b){
    if(a<b)
	return a;
    return b;
}

template<class T> struct index_cmp {
  index_cmp(const T arr) : arr(arr) {}
  bool operator()(const int a, const int b) const
  { return arr[a] > arr[b]; }
  const T arr;
};

int first = 1;

void set_alphas_b0(SVQP2* sv){
  alpha.resize(0);
  b0 = 0;
  if (data_map[UNLABELED].size() > 0){ // add extra column alpha
    alpha.resize(m+1); // alpha of proportion of label constraint
    //     alpha[m]+=sv->x[prob_size-1]; // fill in special alpha
  }else{
    alpha.resize(m); // keeps as many alphas as vectors
  }

  int pos_svs = 0;
  int neg_svs = 0;
  int total_svs = 0;
  // set all alphas to zero
  for(int i=0;i<(int)alpha.size();i++){ alpha[i]=0;}
  // only training vectors may have non-zero alpha
  for(int i=0;i< sv->n;i++){
    alpha[sv->Aperm[i]]+=sv->x[i]; // fill in learnt svs
    if(sv->x[i]>maxa) maxa=(sv->x[i]);
    if(-(sv->x[i])>maxa) maxa=-(sv->x[i]);

    if(Y[data_map[TRAIN][i]] > 0 && sv->x[i] != 0)
      pos_svs++;
    if(Y[data_map[TRAIN][i]] < 0 && sv->x[i] != 0 ) 
      neg_svs++;
    if(sv->x[i] != 0)
      total_svs++;
    //printf("%d %d %lf\n", Y[data_map[TRAIN][i]], Y[data_map[TRAIN][sv->Aperm[i]]], sv->x[i]);
  }
  b0=(sv->gmin+ sv->gmax)/2.0;
  printf("Pos: %d Neg: %d Total: %d\n", pos_svs, neg_svs, total_svs);

  //printf("Maximum alpha=%g\n",maxa);
	
}



void setup_standard_svm_problem(SVQP2* sv)
{
    int fill_level = 0;
  (*sv).verbosity=verbosity;
  (*sv).Afunction=kernel;
  (*sv).Aclosure=(void*) &kparam;
  (*sv).maxcachesize=(long int) 1024*1024*cache_size;
  (*sv).sumflag=use_b0;
  (*sv).epsgr=epsgr;
  for(int j=0;j<(int)data_map[TRAIN].size();++j){
    sv->Aperm[fill_level] = data_map[TRAIN][j];
    double norm = X[data_map[TRAIN][j]]->norm();
    if (Y[data_map[TRAIN][j]] <0){
      int y = -1;
      sv->b[fill_level] = y * (1 + y * (1 - norm)) / norm;
      sv->cmin[fill_level] = - C * norm;
      sv->cmax[fill_level] = 0;
    }else{
	int y = +1;
      sv->b[fill_level]= y * (1 + y * (1 - norm)) / norm;
      sv->cmin[fill_level] = 0;
      sv->cmax[fill_level] = C * norm;
    }
    fill_level++;
  }
}

void get_abs_sel(int& set, int& instance){
    int tmp = instance;
    if(set == 5){
	instance = bag_pointer[data_map[set][instance]]->index;
	data_map[set].erase(data_map[set].begin()+tmp);
    }
    if(set == 5 || set == 7){
	int tmp = instance;
	instance = train_split[instance];
	train_split.erase(train_split.begin()+tmp);
    }
    else{
	int tmp = instance;
	if(tmp<10){
	    instance = train_split[instance];
	    train_split.erase(train_split.begin()+tmp);
	}else{
	    instance = unlab_split[instance];
	    unlab_split.erase(unlab_split.begin()+tmp-10);
	}
    }
}

void print_selections(char * model_file_name){
    char abs_selections[200];
    char selected_ins[200];
    sprintf(selected_ins,"%s-sel",model_file_name);
    sprintf(abs_selections,"%s-abssel",model_file_name);
    ofstream abs_sel(abs_selections);
    ifstream selections(selected_ins);
    while(!selections.eof()){
	int set;
	int instance=-1; 
	selections>>set>>instance;
	if(instance==-1)
	    break;
	get_abs_sel(set,instance);
	abs_sel<<set<<" "<<instance<<endl;
    }
    selections.close();
    abs_sel.close();
}


void load_splits(vector<int>& array, char * split_file){
    ifstream file(split_file);
    while(!file.eof()){
	int tmp;
	file>>tmp;
	array.push_back(tmp);
    }
}

void add_training_data(){
    int count[classes.size()];
    for(int i=0;i<classes.size();i++)
	count[i] = 0;
    
    for(unsigned int i=0;i<classes.size();i++){
	for(unsigned int j=0;j<data_map[UNLABELEDBAGS].size();j++){
	    if(count[i] >=5)
		break;
	    MISparseVector * misv_i = (MISparseVector *)X[data_map[UNLABELEDBAGS][j]];
	    int flag = 0;
	    for(int k=0;k<misv_i->labels.size();k++)
		if(misv_i->labels[k]==classes[i]){
		    flag = 1;
		    break;
		}
	    if(flag){
		int index = data_map[UNLABELEDBAGS][j];
		data_map[UNLABELEDBAGS].erase(data_map[UNLABELEDBAGS].begin()+j);
		j--;
		//count[i]++;
		data_map[TRAINPOS].push_back(index);
		for(int k=0;k<misv_i->labels.size();k++){
		    count[rev_idx[misv_i->labels[k]]]++;
		}
	    }
	}
    }
    
    int count1 = 0;
    for(unsigned int i=0;i<data_map[UNLABELEDBAGS].size();i++){
	MISparseVector * misv_i = (MISparseVector *)X[data_map[UNLABELEDBAGS][i]];
	if(misv_i->label_ != 50)
	    continue;
	int index = data_map[UNLABELEDBAGS][i];
	data_map[UNLABELEDBAGS].erase(data_map[UNLABELEDBAGS].begin()+i);
	i--;
	data_map[TRAINPOS].push_back(index);
	count1++;
	}
	
    
    load_positive_bag_labels();
    printf("Size of TRAINPOS: %d POSINSTANCES: %d\n",data_map[TRAINPOS].size(),data_map[POSINSTANCES].size());
}

void do_initial_training(IncSVM * classifier){
    int set = TRAINPOS;
    int count[classes.size()];
    for(int i=0;i<classes.size();i++)
	count[i] = 0;

    /* add initial_set_size examples and train classifier before running active selection */
    for(int i=0;i<data_map[set].size();i++){
	int index = data_map[set][i];
	MISparseVector * x = (MISparseVector *)(X[index]);
	for(unsigned int lidx=0;lidx<x->labels.size();lidx++){
	    //cout<<x->labels[lidx]<<" "<<rev_idx[x->labels[lidx]]<<endl;
	    if(count[rev_idx[x->labels[lidx]]] < initial_set_size){
		double sum = 0;
			
		MISparseVector * instance = new MISparseVector;
		instance->addInstance(x->getRepresentation());
		
		for(int k=0;k<x->size();k++)
		    x->probs[k] = 1;
		X.push_back(instance);
		Y.push_back(x->labels[lidx]);
		data_map[TRAIN].push_back(X.size()-1);
		add_kernel_row();
		int y = x->labels[lidx];
		double c = C * instance->norm();
		classifier->adddata(X.size()-1,y,c);
		classifier->process_instance(X.size()-1,y);
		x->acquired[lidx] = 1;
		count[rev_idx[x->labels[lidx]]]++;
	    }
	}
	
	if(x->labels.size() != 0 && x->labels[0] == 50)
	    x->labels.clear();
    }
    printf("SIZE OF INITIAL TRAINING SET : %d (%d per class)\n",data_map[TRAIN].size(), initial_set_size);
}



int main(int argc,char**argv)
{
    char input_file_name[1024];
    parse_command_line(argc, argv, input_file_name, model_file_name);
    vector<string> clabels;
    SplitStringUsing(class_labels,",",&clabels);
    
    for(unsigned int i=0;i<clabels.size();i++){
	classes.push_back(atoi(clabels[i].c_str()));
	rev_idx[classes[i]] = i;
    }

    /* loads training split file, validation split file, test split file and the main data file*/
    load_data_file(input_file_name);
    
    /* adds initial_set_size number of examples into the initial training set*/
    add_training_data();
    printf("TRAIN: %d POSINSTANCES: %d UNLABELED: %d X:%d\n",data_map[TRAIN].size(),data_map[POSINSTANCES].size(),data_map[UNLABELEDBAGS].size(),X.size());
    
    /* cache kernel values to speed-up training */
    for(unsigned int i=0;i<X.size();i++){
	vector<double> * tmp = new vector<double>;
	for(unsigned int j=0;j<X.size();j++)
	    tmp->push_back(-1);
	kernel_cache.push_back(tmp);
    }
    
    /* initialize the svm object */
    IncSVM * svm;
    if(multiclass_method == 1)
	svm = new IncSVM_multiclass(classes.size(),&classes);
    else
	svm = new IncSVM_multiclass_cc(classes.size(),&classes);
    
    /* add validation data for fitting the sigmoid to obtain probabilities */
    vector<int> labels;
    for(unsigned int i=0;i<data_map[CROSSVAL].size();i++)
	labels.push_back(Y[data_map[CROSSVAL][i]]);
    svm->add_crossval_data(&data_map[CROSSVAL],&labels);

    /* initializes matrices R, Q (num support vectors, num training examples) */
    svm->initialize(500,1000);

    /* add data into the svm before training */
    load_data_incremental(svm);

    printf("Done loading data\n");
    
    char act_file[200]; // file to store indices of selected examples
    sprintf(act_file,"%s-actsel",model_file_name);

    /* initialize the voi based active selection function */ 
    VOIActiveLearner al(svm, &X, &Y, data_map,&bag_pointer,act_file);
    al.add_labels(&classes);
    al.set_parameters(C,r12,cost_of_labeling, cost_of_labeling*4,3,mi_selection_type);
    printf("Done setting parameters\n");
    
    int count[classes.size()];
    for(unsigned int i=0;i<classes.size();i++)
	count[i] = 0;
    
    C = C / 20;
    printf("C = %lf\n",C);
    
    if(!justtest){
	for(unsigned int i=0;i<X.size();i++)
	    Xsquare.push_back(-1);
		    
	char current_model_file[200];
	sprintf(current_model_file, "%s%d",model_file_name,0);
	sprintf(report_file_name,"%s-stats",current_model_file);
	sprintf(fval_file_name,"%s-out",current_model_file);
	
	do_initial_training(svm);
	sprintf(current_model_file, "%s%d-matlab",model_file_name,0);
	sprintf(report_file_name,"%s-stats",current_model_file);
	sprintf(fval_file_name,"%s-out",current_model_file);
	
	/* variables for calculating and storing different test metrics */
	int selected_set;
	int selected_index;
	int selected_label;
	double cost;
	double total_cost = 0;
	char acc_file[200];
	char auroc_per_class_file[200];
	char auroc_file[200];
	char auroc_per_class_file1[200];
	char auroc_file1[200];
	char risk_file[200];
	char imgfilename[200];
	sprintf(report_file_name,"%s-stats",model_file_name);
	sprintf(imgfilename,"%s-selections",model_file_name);
	sprintf(acc_file,"%s-acc",model_file_name);
	sprintf(auroc_file,"%s-auroc",model_file_name);
	sprintf(risk_file,"%s-risk",model_file_name);
	sprintf(auroc_per_class_file,"%s-auroc-class",model_file_name);
	sprintf(auroc_file1,"%s-region-auroc",model_file_name);
	sprintf(auroc_per_class_file1,"%s-region-auroc-class",model_file_name);
	sprintf(fval_file_name,"%s-out",model_file_name);
	ofstream acc_out(acc_file);
	ofstream risk_out(risk_file);
	ofstream auroc_out(auroc_file);
	ofstream auroc_per_class_out(auroc_per_class_file);
	ofstream auroc_out1(auroc_file1);
	ofstream auroc_per_class_out1(auroc_per_class_file1);
	ofstream actsel(act_file);
	ofstream selout(imgfilename);
	double area[classes.size()];
	double area1[classes.size()];
	char imgname[20];
	int actual_index;

	/* **************************************************************************** 
	               The main training loop using active/random selection. 
	*******************************************************************************/
	for(int iter = 1;iter<active_iterations;iter++){
	    sprintf(current_model_file, "%s%d",model_file_name,iter-1);
	    sprintf(fval_file_name,"%s-out",current_model_file);

	    /* test the current model */
	    double acc = test_model(svm,TMP); // multi-class accuracy 
	    auroc_per_class_out<<total_cost;
	    double auroc = test_label_presence1(svm,TEST,area); // auroc for label presence (bag level prediction)
	    double auroc1 = test_region_accuracy(svm,TMP,area1); // auroc for instance(region) level prediction
	    double risk = calculate_risk(svm,TMP);

	    /* save all stats */
	    for(unsigned int i=0;i<classes.size();i++){
		double tmp = area[i];
		auroc_per_class_out<<" "<<tmp;
		auroc_per_class_out1<<" "<<area1[i];
	    }
	    auroc_per_class_out<<endl;
	    auroc_per_class_out1<<endl;
	    acc_out<<total_cost<<" "<<acc<<endl;
	    auroc_out<<total_cost<<" "<<auroc<<endl;
	    auroc_out1<<total_cost<<" "<<auroc1<<endl;
	    risk_out<<total_cost<<" "<<risk<<endl;
	    printf("Average AUROC = %f\n",auroc);
	    printf("Average AUROC(region) = %f\n",auroc1);
	    printf("------------------- End of Iteration %d -----------------------------------\n",iter-1);
	    printf("******************* Iteration %d(%d) **********************************\n",iter,active_iterations);
	    
	    /* run the active learner once to add a single example and train */
	    cost = al.run_once(selected_index, selected_set,selected_label,imgname,actual_index);
	    if(cost==-1){
		printf("End of iterations\n");
		break;
	    }

	    /* save details of selected example */
	    actsel<<selected_set<<" "<<selected_index<<" "<<cost<<endl;
	    if(selected_set != 7)
		selout<<selected_set<<" "<<selected_label<<endl;
	    else{
		selout<<selected_set;
		MISparseVector * misv_i = (MISparseVector*)X[actual_index];
		for(int i=0;i<misv_i->labels.size();i++)
		    selout<<" "<<misv_i->labels[i];
		selout<<endl;
	    }

	    /* total annotation cost spent till now */
	    total_cost += cost;
	}
	selout.close();
	actsel.close();
	acc_out.close();
	auroc_out.close();
	auroc_per_class_out.close();
	auroc_out1.close();
	auroc_per_class_out1.close();
	//delete svm;
    }
    else{
	alpha.resize(0);
	// set all alpha's of nonSV to zero
	for (int i = 0; i != m; ++i)
	    alpha.push_back(0.0); 
	int msz = load_model(model_file_name);
	m+=msz;
	for(unsigned int i=0;i<X.size();i++)
	    Xsquare.push_back(-1);
	//test_current_model(svm,0,TEST);
    }
    return 0;
}

