#include "libincsvm.h"
#include <vector>
#include <limits>
#include <math.h>

#define Malloc(type,n) (type *)malloc((n)*sizeof(type))


extern "C"{
#include <cblas.h>
}

using namespace std;

#define MARGIN 0
#define ERROR  1
#define RESERVE 2
#define UNLEARNED 3


double kernel(int i,int j,void * params);

void print_vector(vector<double>& a){
  for(unsigned int i=0;i<a.size();i++)
    //if(a[i]!=0)
      printf("%g ",a[i]);
  printf("\n");
}


void print_vector(vector<int>& a){
  for(unsigned int i=0;i<a.size();i++)
    printf("%d ",a[i]);
  printf("\n");
}

void print_matrix(vector<vector<double>*>& mat){
  for(unsigned int i=0;i<mat.size();i++)
    print_vector(*mat[i]);
}

int IncSVM::find(int indc, vector<int>& indices){
  for(unsigned int i=0;i<indices.size();i++)
    if(indices[i] == indc)
      return i;
  return -1;
}

double IncSVM::svmeval(int indc,vector<double>* Qc){
  double fx = b;
  int sets[] = {MARGIN,ERROR,RESERVE};//,UNLEARNED};
  for(int j=0;j<3;j++){
    int set = sets[j];
    for(unsigned int i=0;i<ind[set].size();i++){
      int index = ind[set][i];
      double k = kernel(indc,index,NULL);
      if(Qc!=NULL){
	Qc[set].push_back(k*y[index]*y[indc]);
      }
      my_assert(index<(int)a.size() && index<(int)y.size(),"libsvm: Line 72");
      if(set==MARGIN || set==ERROR || a[index] >0)
	fx = fx + a[index]*y[index]*k;
    }
  }
  return fx;
}


void IncSVM::move_ind(int cstatus, int nstatus, int index){
  my_assert(index<(int)ind[cstatus].size(),"libsvm: Line 81");
  int indc = ind[cstatus][index];
  ind[cstatus].erase(ind[cstatus].begin()+index);
  ind[nstatus].push_back(indc);
}

int IncSVM::move_indr(int cstatus, int indc){
  int removed_i = -1;
  move_ind(cstatus,RESERVE,indc);
  // implement max reserve vectors thing
  // (optional)
  return removed_i;
}

void IncSVM::bookkeeping(int indss, int cstatus, int nstatus,int& indco, int& idx){
  indco = -1;
  idx = -1;
  if(cstatus == nstatus)
    return;
  switch(nstatus){
  case RESERVE:
    a[indss] = 0;
    break;
  case ERROR:
    a[indss] = c[indss];
    break;
  }
  
  my_assert(indss<(int)a.size() && indss<(int)c.size(),"libsvm: Line 117");
  int index = find(indss,ind[cstatus]);
  my_assert(index != -1,"libsvm: 119");
  if(cstatus == MARGIN)
    indco = index+1;
  
  if(nstatus == RESERVE)
    idx = move_indr(cstatus,index);
  else
    move_ind(cstatus,nstatus,index);
}

void vector_plus_vector(int start1, int c1, vector<double>& v1, int start2, int c2, vector<double>& v2, int start_res, vector<double>& result, int size){
  for(int i=0;i<size;i++)
    result[start_res+i] = c1*v1[start1+i]+c2*v2[start2+i];
}

void vector_plus_vector(int start1, int c1, vector<double>& v1, int start2, int c2, vector<double>& v2, vector<int>& indices, vector<double>& result){
  for(unsigned int i=0;i<indices.size();i++)
    result[indices[i]] = c1*v1[start1+i]+c2*v2[start2+i];
}


void vector_plus_vector(vector<int>& ind1, int c1, vector<double>& v1, int start2, int c2, vector<double>& v2, vector<int>& indres, vector<double>& result){
  char msg[200];
  sprintf(msg,"libsvm: Line 121 sizes %d %d",ind1.size(),indres.size());
  my_assert(ind1.size()==indres.size(),msg);
  for(unsigned int i=0;i<ind1.size();i++)
    result[indres[i]] = c1*v1[ind1[i]]+c2*v2[start2+i];
}

double vector_star_vector(vector<double>& v1, vector<double>& v2){
  char msg[200];
  sprintf(msg,"libsvm: Line 147 sizes %d %d",v1.size(),v2.size());
  my_assert(v1.size()==v2.size(),msg);
  double res = 0;
  for(unsigned int i=0;i<v1.size();i++)
    res += v1[i]*v2[i];
  return res;
}

double vector_star_vector(double v1[], vector<double>& v2){
  double res = 0;
  for(unsigned int i=0;i<v2.size();i++)
    res += v1[i]*v2[i];
  return res;
}


double vector_star_vector(double v1[], double v2[], int size){
  double res = 0;
  for(int i=0;i<size;i++)
    res += v1[i]*v2[i];
  return res;
}


void matrix_star_vector(int c, vector<vector<double>*>& A,vector<double>& x,vector<double>& res){
  res.clear();
  for(unsigned int i=0;i<A.size();i++)
    res.push_back(c*vector_star_vector(*(A[i]),x));
}

void matrix_star_vector(int c, vector<vector<double>*>& A,double x[],vector<double>& res){
  res.clear();
  for(unsigned int i=0;i<A.size();i++)
    res.push_back(c*vector_star_vector(x,*(A[i])));
}

void matrix_star_vector(int c, Matrix& A,vector<double>& x,vector<double>& res){
  res.clear();
  for(int i=0;i<A.size;i++)
    res.push_back(c*vector_star_vector(A.array[i],x));
}

void matrix_star_vector(int c, Matrix& A,double x[],vector<double>& res){
  res.clear();
  for(int i=0;i<A.size;i++)
    res.push_back(c*vector_star_vector(x,A.array[i],A.row_size[i]));
}



void matrix_star_vector(int c, vector<double>** A,int size, vector<double>& x,vector<double>& res){
  res.clear();
  for(int i=0;i<size;i++)
    res.push_back(c*vector_star_vector(*(A[i]),x));
}




void IncSVM::updateRQ(int indc){
  int rows = _Rs.size;
  stopwatch sw;
  if(rows>2){
    double val = _Rs.array[indc][indc];
    for(int i=0;i<rows;i++)    
      for(int j=0;j<rows;j++)
	if(i != indc && j != indc)
	  _Rs.array[i][j] = _Rs.array[i][j] -  _Rs.array[i][indc]*_Rs.array[indc][j]/val;
    _Rs.del_col(indc);
    _Rs.del_row(indc);
  }
  else{
    my_assert(rows==2,"libsvm: Line 184");
    _Rs.row_size[0] = 0;
    _Rs.row_size[1] = 0;
    _Rs.size = 0;
  }
  _Q.del_col(indc);
}

void IncSVM::updateRQ(vector<double>& beta, vector<double>& tmp1, int indc){
  stopwatch watch;
  double gamma = tmp1[0];
  int rows = _Rs.size;
  if(gamma <deps)
    gamma = deps;
  if(rows >1){
    for(int i=0;i<_Rs.size;i++){
      for(int j=0;j<_Rs.row_size[i];j++)
	_Rs.array[i][j] = _Rs.array[i][j] + beta[i]*beta[j]/gamma;
      _Rs.push_back(i,beta[i]/gamma);
    }
    _Rs.addrow();
    for(unsigned int i=0;i<beta.size();i++)
      _Rs.push_back(beta[i]/gamma);
    _Rs.push_back(1/gamma);
  }
  else{
    _Rs.addrow();
    _Rs.push_back(-(kernel(indc,indc,NULL)+deps));
    _Rs.push_back(y[indc]);
    _Rs.addrow();
    _Rs.push_back(y[indc]);
    _Rs.push_back(0);
  }
  for(unsigned int i=0;i<y.size();i++){
    double val = 0;
    if(i==(unsigned int)indc)
      val = deps;
    if(_Q.get_row_size(i)>0)
      _Q.push_back(i,y[indc]*y[i]*kernel(indc,i,NULL)+val);
  }
}

void IncSVM::min(double * a,int size,double& res, int& ind){
  res = std::numeric_limits<double>::infinity();
  ind = 0;
  for(int i=0;i<size;i++)
    if(a[i]<res){
      res = a[i];
      ind = i;
    }
}

double IncSVM::abs(double val){
  if(val<0)
    val = -val;
  return val;
}


void IncSVM::min_delta(vector<int>& flags, vector<double> psi_initial, vector<double> psi_final,vector<double> psi_sens,double& min_d, int& k){
  vector<int> ind;
  int idx;
  for(unsigned int i=0;i<flags.size();i++)
    if(flags[i])
      ind.push_back(i);
  if(ind.size() >0){
    double deltas[ind.size()];
    for(unsigned int i=0;i<ind.size();i++){
      deltas[i] = (psi_final[ind[i]] - psi_initial[ind[i]])/psi_sens[ind[i]];
    }
    min(deltas,ind.size(),min_d,idx);
    k = ind[idx];
    double max_sens = -std::numeric_limits<double>::infinity();
    for(unsigned int i=0;i<ind.size();i++){
      if(deltas[i] == min_d && abs((double)psi_sens[ind[i]])>max_sens){
	max_sens = abs((double)psi_sens[ind[i]]);
	k = ind[i];
      }
    }
  }
  else{
    min_d = std::numeric_limits<double>::infinity();
    k = -1;
  }
}




void IncSVM::select_indices(vector<double>& array, vector<int>& indices, vector<double>& res){
  for(unsigned int i=0;i<indices.size();i++){
    my_assert((unsigned int)indices[i]<array.size(),"libsvm: Line 305");
    res.push_back(array[indices[i]]);
  }
}

void IncSVM::min_delta_acb(int indc,vector<double>& gamma,vector<double>& beta,int polc,int rflag, double& min_dacb, int& indss_,int& cstatus_,int& nstatus_){
  vector<int> indss(6,0);
  indss[0]=indss[1]=indss[2] = indc;
  vector<int> cstatus(6,0);
  vector<int> nstatus(6,0);
  double delta_m,delta_e,delta_r;
  my_assert((unsigned int)indc<g.size()&&(unsigned)indc<gamma.size(),"libsvm: Line 316");
  my_assert((unsigned int)indc<c.size()&&(unsigned)indc<a.size(),"libsvm: Line 317");
  if(polc==1){
    delta_m = -g[indc]/gamma[indc];
    nstatus[0] = MARGIN;
    if(beta.size()>1){
      delta_e = c[indc] - a[indc];
      nstatus[1] = ERROR;
    }
    else
      delta_e = std::numeric_limits<double>::infinity();
    delta_r = std::numeric_limits<double>::infinity();
  }
  else{
    if(g[indc]>0){
      delta_m = g[indc]/gamma[indc];
      nstatus[0] = MARGIN;
    }
    else
      delta_m = std::numeric_limits<double>::infinity();
    
    if(a[indc] <= c[indc]){
      delta_e = std::numeric_limits<double>::infinity();
      delta_r = a[indc];
      if(g[indc]>0)
	nstatus[2] = RESERVE;
      else
	nstatus[2] = UNLEARNED;
    }
    else{
      delta_e  = a[indc] - c[indc];
      delta_r = std::numeric_limits<double>::infinity();
      nstatus[1] = ERROR;
    }
  }


  // change in a_c or b that causes a margin vector to an error or reserve vector 
  vector<double> beta_s;
  vector<int> flags;
  double delta_mer;
  int idx;
  if(beta.size()>1){
    for(unsigned int i=1;i<beta.size();i++){
      beta_s.push_back(polc*beta[i]);
      if(abs(polc*beta[i])>0)
	flags.push_back(1);
      else
	flags.push_back(0);
    }
    
    vector<double> a_tmp;
    vector<double> c_tmp;
  
    select_indices(a,ind[MARGIN],a_tmp);
    select_indices(c,ind[MARGIN],c_tmp);
    for(unsigned int i=0;i<c_tmp.size();i++)
      c_tmp[i] *= (beta_s[i]>0);
    min_delta(flags,a_tmp,c_tmp,beta_s,delta_mer,idx);
    if(delta_mer < std::numeric_limits<double>::infinity()){
      indss[3] = ind[MARGIN][idx];
      cstatus[3] = MARGIN;
      nstatus[3] = ERROR *(beta_s[idx]>0)+RESERVE*(beta_s[idx]<0);
    }
  }
  else
    delta_mer = std::numeric_limits<double>::infinity();

  // change in a_c or b that causes an error vector to change to a margin vector
  vector<double> gamma_e;
  flags.clear();
  double delta_em;
  select_indices(gamma,ind[ERROR],gamma_e);
  for(unsigned int i=0;i<gamma_e.size();i++){
    gamma_e[i] *= polc;
    flags.push_back(gamma_e[i]>0);
  }
  vector<double> g_tmp;
  vector<double> z_tmp(ind[ERROR].size(),0);
  select_indices(g,ind[ERROR],g_tmp);
  min_delta(flags,g_tmp,z_tmp,gamma_e,delta_em,idx);
  if(delta_em < std::numeric_limits<double>::infinity()){
    indss[4] = ind[ERROR][idx];
    cstatus[4] = ERROR;
    nstatus[4] = MARGIN;
  }
  

  // change in a_c or b that causes a reserve vector to change to a margin vector
  double delta_rm;
  if(rflag){
    vector<double> gamma_r;
    vector<int> flags;
    vector<double> g_tmp;
    vector<double> z_tmp(ind[RESERVE].size(),0);
    select_indices(g,ind[RESERVE],g_tmp);
    select_indices(gamma,ind[RESERVE],gamma_r);
    for(unsigned int i=0;i<gamma_r.size();i++){
      gamma_r[i] *= polc;
      my_assert(i<g_tmp.size(),"libsvm: Line 415");
      flags.push_back(g_tmp[i] >=0 && gamma_r[i] <0);
    }
    min_delta(flags,g_tmp,z_tmp,gamma_r,delta_rm,idx);
    if(delta_rm < std::numeric_limits<double>::infinity()){
      indss[5] = ind[RESERVE][idx];
      cstatus[5] = RESERVE;
      nstatus[5] = MARGIN;
    }
  }
  else
    delta_rm = std::numeric_limits<double>::infinity();
  
  double tmp[] = {delta_m,delta_e,delta_r,delta_mer,delta_em,delta_rm};
  int min_ind;
  min(tmp,6,min_dacb,min_ind);
  indss_ = indss[min_ind];
  cstatus_ = cstatus[min_ind];
  nstatus_ = nstatus[min_ind];
  min_dacb = polc*min_dacb;
}


void IncSVM::initialize(int _m, int _n){
  _Rs.initialize(_m,_m);
  _Q.initialize(_n,_m);
  deps = 1e-3;
  max_reserve_vectors = 3000;
  b = 0;
  kernel_evals = 0;
  perturbations = 0;
}



void IncSVM::adddata(int indc, int y_new, double c_new){
  a.push_back(0);
  c.push_back(c_new);
  g.push_back(0);
  //_Q.addrow();
  y.push_back(y_new);
  ind[UNLEARNED].push_back(indc);
}


int IncSVM::learn(int indc, int rflag){
  char msg[200];

  sprintf(msg,"libsvm: Line 478, indc %d Q %d y %d g %d a %d\n",indc,_Q.size,y.size(),g.size(),a.size());
  my_assert((unsigned)indc<y.size()&&(unsigned)indc<g.size()&&(unsigned)indc<a.size(),msg);
  _Q.addrow(indc);
  _Q.push_back(indc,y[indc]);
  for(unsigned int i=0;i<ind[MARGIN].size();i++){
    my_assert(ind[MARGIN][i]<(int)y.size(),"libsvm: Line 481");
    _Q.push_back(indc,y[ind[MARGIN][i]]*y[indc]*kernel(indc,ind[MARGIN][i],NULL));
    //if(indc == 197)
    //printf("i:%d value is %lf\n",i+1,_Q.array[indc][i+1]);
  }
  stopwatch sw;
  sw.reset();
  vector<double> Qc[4];
  double fx = svmeval(indc,Qc);
  g[indc] = y[indc]*fx-1;
  if(g[indc]>=0){
    int indco,idx;
    bookkeeping(indc,UNLEARNED,RESERVE,indco,idx);
    //if(indc >=130 && indc <=150)
    //printf("After learning %d is in MARGIN:%d, ERROR: %d, RESERVE: %d, g[indc] = %lf\n",indc,find(indc,ind[MARGIN]),find(indc,ind[ERROR]),find(indc,ind[RESERVE]),g[indc]);
    return RESERVE;
  }
  double Qcc = kernel(indc,indc,NULL)+deps;
  int converged=0;
  int cstatus,nstatus;
  double min_delta_param;
  int num_MVs = ind[MARGIN].size();
  stopwatch s1;

  while(!converged){
    s1.reset();
    perturbations++;
    vector<double> beta;
    vector<double> gamma(y.size(),0);
    if(num_MVs>0){
      //calculate beta
      for(int i=0;i<_Rs.size;i++){
	double val = _Rs.array[i][0] * y[indc] + vector_star_vector(_Rs.array[i]+1, Qc[MARGIN]);
	beta.push_back(-val);
      }
      
      //calculate gamma
      int sets[] = {ERROR, RESERVE};
      for(int i=0;i<2;i++)
	for(unsigned int j=0;j<ind[sets[i]].size();j++){
	  my_assert(ind[sets[i]][j]<(int)gamma.size(),"libsvm: Line 551");
	  double gamma_tmp = vector_star_vector(_Q.get_row(ind[sets[i]][j]), beta);
	  gamma[ind[sets[i]][j]] = gamma_tmp + Qc[sets[i]][j];
	}
      double gamma_tmp = vector_star_vector(_Q.get_row(indc), beta);
      gamma[indc] = gamma_tmp+Qcc;
      if(gamma[indc]<0){
	print_vector(beta);
	_Q.print_row(indc);
	printf("indc = %d gamma_tmp = %lf Qcc = %lf\n",indc,gamma_tmp,Qcc);
	printf("Kernel val is %lf with ind %d\n",kernel(indc,ind[MARGIN][1],NULL),ind[MARGIN][1]);
	fprintf(stderr,"Kernel is not positive\n");
	exit(21);
      }
    }
    else{
      beta.push_back(y[indc]);
      for(unsigned int i=0;i<y.size();i++)
	gamma[i] = y[indc]*y[i];
    }

    int indss;
    watch.reset();
    min_delta_acb(indc,gamma,beta,1,rflag,min_delta_param,indss,cstatus,nstatus);
    watch.reset();
    if(num_MVs>0){
      a[indc] += min_delta_param;
      for(unsigned int i=0;i<ind[MARGIN].size();i++)
	a[ind[MARGIN][i]] += beta[i+1]*min_delta_param;
    }
    b += beta[0]*min_delta_param;
    for(unsigned int i=0;i<g.size();i++)
      g[i] += min_delta_param*gamma[i];
    converged = (indss == indc);
    if(converged){
      cstatus = UNLEARNED;
      Qc[nstatus].push_back(Qcc);
    }
    else{
      int ind_temp = find(indss,ind[cstatus]);
      Qc[nstatus].push_back(Qc[cstatus][ind_temp]);
      Qc[cstatus].erase(Qc[cstatus].begin()+ind_temp);
    }
    int indco,idx;
    bookkeeping(indss,cstatus,nstatus,indco,idx);
    //remove reserve vectors
    for(unsigned int i=0;i<ind[MARGIN].size();i++)
      g[ind[MARGIN][i]] = 0;
    watch.reset();
    if(nstatus == MARGIN){
      num_MVs++;
      if(num_MVs>1){
	double gamma1;
	if(converged)
	   gamma1 = gamma[indss];
      	else{
	  matrix_star_vector(-1,_Rs,_Q.get_row(indss),beta);
	  gamma1 = kernel(indss,indss,NULL)+deps+vector_star_vector(_Q.get_row(indss),beta);
	}
	gamma.clear();
	gamma.push_back(gamma1);
      }
      updateRQ(beta,gamma,indss);
     }
    else if(cstatus == MARGIN){
      num_MVs--;
      updateRQ(indco);
     }
  }
  //if(indc >=130 && indc <=150){
  //for(int i = 130;i<=150;i++)
  //  printf("After learning %d, %d is in MARGIN:%d, ERROR: %d, RESERVE: %d, g[indc] = %lf\n",indc,i,find(i,ind[MARGIN]),find(i,ind[ERROR]),find(i,ind[RESERVE]),g[i]);
  //}
  return nstatus;
}


int IncSVM::unlearn(int indc){
  int num_MVs = ind[MARGIN].size();
  char msg[200];
  sprintf(msg,"libsvm: Line 648, indc %d g %d y %d Q %d a %d\n",indc,g.size(),y.size(),_Q.size,a.size());
  my_assert((unsigned)indc<g.size()&&(unsigned)indc<y.size()&&(unsigned)indc<a.size(),msg);
  if(g[indc]<0){
    int idx = find(indc,ind[ERROR]);
    sprintf(msg,"libsvm: Line 649 idx = %d",indc);
    if(idx == -1){
      int a = find(indc,ind[MARGIN]);
      int b = find(indc,ind[RESERVE]);
      int c = find(indc,ind[UNLEARNED]);
      printf("Index %d not found in ERROR, MARGIN:%d, RESERVE:%d, UNLEARNED: %d\n",indc,a,b,c);
    }
    my_assert(idx!=-1,msg);
    ind[ERROR].erase(ind[ERROR].begin()+idx);
  }
  else if(g[indc] == 0){
    int idx = find(indc,ind[MARGIN]);
    if(idx != -1){
      ind[MARGIN].erase(ind[MARGIN].begin()+idx);
      num_MVs --;
      updateRQ(idx+1);
    }
    else{
      idx = find(indc,ind[RESERVE]);
      my_assert(idx!=-1,"libsvm: Line 664");
      ind[RESERVE].erase(ind[RESERVE].begin()+idx);
    }
  }
  else{
    int idx = find(indc,ind[RESERVE]);
    my_assert(idx!=-1,"libsvm: Line 670");
    ind[RESERVE].erase(ind[RESERVE].begin()+idx);
    ind[UNLEARNED].push_back(indc);
    _Q.clear(indc);
    return RESERVE;
  }

  vector<double> Qc[4];
  svmeval(indc,Qc);
  double Qcc = kernel(indc,indc,NULL)+deps;
  int cstatus,nstatus;
  double min_delta_param;
  int converged = 0;
  while(!converged){
    vector<double> beta;
    vector<double> gamma(y.size(),0);
    if(num_MVs>0){
      vector<vector<double>*> Q_tmp;
      //calculate beta
      for(int i=0;i<_Rs.size;i++){
	double val = _Rs.array[i][0] * y[indc] + vector_star_vector(_Rs.array[i]+1, Qc[MARGIN]);
	beta.push_back(-val);
      }

      //calculate gamma
      int sets[] = {ERROR, RESERVE};//,UNLEARNED};
      for(int i=0;i<2;i++){
	for(unsigned int j=0;j<ind[sets[i]].size();j++){
	  double gamma_tmp = vector_star_vector(_Q.get_row(ind[sets[i]][j]), beta);
	  gamma[ind[sets[i]][j]] = gamma_tmp + Qc[sets[i]][j];
	}
      }
      double gamma_tmp = vector_star_vector(_Q.get_row(indc), beta);
      gamma[indc] = gamma_tmp+Qcc;
    }
    else{
      beta.push_back(y[indc]);
      for(unsigned int i=0;i<y.size();i++)
	gamma[i] = y[indc]*y[i];
    }
    int indss;
    min_delta_acb(indc,gamma,beta,-1,1,min_delta_param,indss,cstatus,nstatus);
    converged = (indss == indc);
    if(num_MVs>0){
      a[indc] += min_delta_param;
      for(unsigned int i=0;i<ind[MARGIN].size();i++)
	a[ind[MARGIN][i]] += beta[i+1]*min_delta_param;
    }
    b += beta[0]*min_delta_param;
    for(unsigned int i=0;i<g.size();i++)
      g[i] += min_delta_param*gamma[i];
    if(!converged){
      int ind_temp = find(indss,ind[cstatus]);
      Qc[nstatus].push_back(Qc[cstatus][ind_temp]);
      Qc[cstatus].erase(Qc[cstatus].begin()+ind_temp);
      int indco,idx;
      bookkeeping(indss,cstatus,nstatus,indco,idx);
      //reserve vectors thingy
      if(nstatus==MARGIN){
	num_MVs++;
	if(num_MVs > 1){
	  matrix_star_vector(-1,_Rs,_Q.get_row(indss),beta);
	  double gamma1 = kernel(indss,indss,NULL)+deps+vector_star_vector(_Q.get_row(indss),beta);
	  gamma.clear();
	  gamma.push_back(gamma1);
	}
	updateRQ(beta,gamma,indss);
      }
      else if(cstatus == MARGIN){
	num_MVs --;
	updateRQ(indco);
      }
    }
    else{
      ind[nstatus].push_back(indc);
      if(nstatus == MARGIN){
	num_MVs++;
	if(num_MVs > 1){
	  matrix_star_vector(-1,_Rs,_Q.get_row(indss),beta);
	  double gamma1 = kernel(indss,indss,NULL)+deps+vector_star_vector(_Q.get_row(indss),beta);
	  gamma.clear();
	  gamma.push_back(gamma1);
	}
	updateRQ(beta,gamma,indss);
      }
    }
    for(unsigned int i=0;i<ind[MARGIN].size();i++)
      g[ind[MARGIN][i]] = 0;
  }
  _Q.clear(indc);
  return MARGIN;
}


double IncSVM::get_w2()
{
  unsigned int i;
  double s = 0;
  for (i=0; i<a.size(); i++){
    if(a[i]<=0)
      continue;
    if (y[i]>0)
      s += a[i] * ( g[i] + 1.0 );
    else 
      s += y[i] * a[i] * ( g[i] - 1.0 );
  }
  return s/2.0;
}


vector<double>& IncSVM::get_model(double& b_){
  b_ = b;
  return a;
}


void IncSVM::setY(int idx, int y_new){
  int y_old = (int)y[idx];
  y[idx] = y_new;
  char msg[200];
  sprintf(msg,"libincsvm: Line 709 sizes %d %d\n",idx,_Q.size);
  for(int i=0;i<_Q.get_row_size(idx);i++)
    _Q.get_row(idx)[i] *= y_old*y_new;
}


void IncSVM_posterior::process_instance(int idx, int label){
  setY(idx,label);
  learn(idx,1);
  fit_sigmoid(100);
}


void IncSVM_posterior::unlearn_instance(int idx, int label){
  unlearn(idx);
  fit_sigmoid(100);
}

void IncSVM_posterior::process_instance_only(int idx, int label){
  setY(idx,label);
  learn(idx,1);
}


void IncSVM_posterior::unlearn_instance_only(int idx, int label){
  unlearn(idx);
}

double IncSVM_posterior::svmprob(int indc, int label){
  double val = svmeval(indc,NULL);
  double prob = 1/(1+exp(sigmoidA*val+sigmoidB));
  if(label == -1)
    prob = 1 - prob;
  return prob;
}


void IncSVM_posterior::svmprobs(int indc, double * probs){
  probs[0] = svmprob(indc,1);
  probs[1] = svmprob(indc,-1);
}

int IncSVM_posterior::classify(int indc){
  double prob = svmprob(indc,1);
  if(prob>=0.5)
    return 1;
  return -1;
}

void IncSVM_posterior::fit_sigmoid(int max_iter){
    double prior1=0,prior0=0;
    double out[crossval_size];

    vector<int>& indices = *crossval_indices;
    vector<int>& c_labels = *crossval_labels;
    
    for(int i=0;i<crossval_size;i++){
      out[i] = IncSVM::svmeval(indices[i],NULL);
      if(c_labels[i]==1)
	prior1 ++;
      else if(c_labels[i]==-1)
	prior0 ++;
    }

  sigmoidA = 0;
  sigmoidB = log((prior0+1)/(prior1+1));
  int size = (int)(prior0+prior1);
  double loTarget = 1/(prior0+2);
  double hiTarget = (prior1+1)/(prior1+2);
  double lambda = 0.001;
  double olderr = 1e300,err;
  double t=0,d1,d2,oldA,oldB,det;

  for(int i=0;i<size;i++){
    pp[i] = (prior1+1)/(prior0+prior1+2);
  }
  int count = 0;
  int num_iter = 0;
  for(int it=0;it<max_iter;it++){
    double a=0,b=0,c=0,d=0,e=0;
    num_iter++;
    for(int i=0;i<size;i++){
      
      if(c_labels[i]==1)
	t = hiTarget;
      else
	t = loTarget;

      d1 = pp[i] - t;
      d2 = pp[i] * (1-pp[i]);
      a += out[i]*out[i]*d2;
      b += d2;
      c += out[i]*d2;
      d += out[i]*d1;
      e += d1;
    }
    
    if(fabs(d) < 1e-9 && fabs(e) < 1e-9)
      break;
    
    oldA = sigmoidA;
    oldB = sigmoidB;
    err = 0;

    while(1){
      det = (a+lambda)*(b+lambda) - c*c;
      if(det == 0){
	lambda *= 10;
	continue;
      }
      sigmoidA = oldA + ((b+lambda)*d-c*e)/det;
      sigmoidB = oldB + ((a+lambda)*e-c*d)/det;
      err = 0;
      for(int i=0;i<size;i++){
	double p = 1/(1+exp(out[i]*sigmoidA+sigmoidB));
	pp[i] = p;
	double l1 = -200,l2=-200;
	if(p>0)
	    l1 = log(p);
	if(1-p>0)
	    l2 = log(1-p);
	
	err -= t*l1 + (1-t)*l2;
      }
      
      if(err < olderr*(1+1e-7)){
	lambda *= 0.1;
	break;
      }

      lambda *= 10;
      if(lambda > 1e6)
	break;
    }

    double diff = err -  olderr;
    double scale = 0.5*(err+olderr+1);
    if(diff > -1e-3*scale && diff < 1e-7*scale)
      count++;
    else
      count = 0;
    olderr = err;
    if(count==3)
      break;
  }
}

  
  
void IncSVM_multiclass::add_crossval_data(vector<int>* crossval_indices, vector<int> * crossval_labels){
  for(int i=0;i<num_classes;i++)
    for(int j=i+1;j<num_classes;j++){
      vector<int> * indices = new vector<int>;
      vector<int> * c_labels = new vector<int>;
      
      for(unsigned int k=0;k<crossval_indices->size();k++)
	if((*crossval_labels)[k] == (*labels)[i] || (*crossval_labels)[k] == (*labels)[j]){
	  indices->push_back((*crossval_indices)[k]);
	  if((*crossval_labels)[k] == (*labels)[i])
	    c_labels->push_back(1);
	  else
	    c_labels->push_back(-1);
	}
      classifiers[i][j]->add_crossval_data(indices,c_labels);
    }

  vector<int> * indices = new vector<int>;
  vector<int> * c_labels = new vector<int>;
  int numpos = 0;
  int numneg = 0;

  for(unsigned int k=0;k<crossval_indices->size();k++){
    int label = (*crossval_labels)[k];
    int l_idx = -1;
    indices->push_back((*crossval_indices)[k]);
    for(int i=0;i<num_classes;i++){
      if((*labels)[i] == label){
	l_idx = i;
	break;
      }
    }
    if(l_idx != -1){
      c_labels->push_back(1);
      numpos++;
    }
    else{
      c_labels->push_back(-1);
      numneg++;
    }
  }
  printf("Cross val for classes vs background %d:%d\n",numpos,numneg);
  cl->add_crossval_data(indices,c_labels);
}

double IncSVM_multiclass::svmprob(int indc, int label){
  double p[labels->size()];
  svmprobs(indc,p);
  int idx = -1;
  for(unsigned int i=0;i<labels->size();i++)
    if((*labels)[i] == label){
      idx = i;
      break;
    }
  return p[idx];
}

double IncSVM_multiclass::svmprob_ispresent(vector<int>* indc, int label){
  double p[labels->size()];
  svmprobs_ispresent(indc,p);
  int idx = -1;
  for(unsigned int i=0;i<labels->size();i++)
    if((*labels)[i] == label){
      idx = i;
      break;
    }
  return p[idx];
}


void IncSVM_multiclass::svmprobs_ispresent(vector<int> * indc, double * probs){
  double *tmp_prob[indc->size()];
  for(unsigned int i=0;i<indc->size();i++){
    tmp_prob[i] = new double[num_classes];
    svmprobs((*indc)[i], tmp_prob[i]);
  }

  for(int i=0;i<num_classes;i++){
    double prob = 1;
    double max = 0;
    for(unsigned int j=0;j<indc->size();j++){
      prob *= (1-tmp_prob[j][i]);
      if(tmp_prob[j][i] > max)
	max = tmp_prob[j][i];
    }
    probs[i] = max;
  }
  
  for(unsigned int i=0;i<indc->size();i++)
    delete tmp_prob[i];
}



void IncSVM_multiclass::svmprobs(int indc, double * probs){
  double **r = new double*[num_classes];
  for(int i=0;i<num_classes;i++)
    r[i] = new double[num_classes];

  for(int i=0;i<num_classes;i++)
    for(int j=i+1;j<num_classes;j++){
      if(i==j)
	continue;
      else if(i<j){
	r[i][j] = classifiers[i][j]->svmprob(indc,1);
	r[j][i] = 1 - r[i][j];
      }
      else
	r[i][j] = classifiers[j][i]->svmprob(indc,-1);
    }
  estimate_probabilities(r,probs);

  for(int i=0;i<num_classes;i++){
    delete r[i];
  }
  delete r;
}


void IncSVM_multiclass::adddata(int idx, int y, double C){
  int l_idx = -1;
  for(int i=0;i<num_classes;i++){
    if((*labels)[i] == y){
      l_idx = i;
      break;
    }
  }
  
  for(int i=0;i<num_classes;i++)
    for(int j=i+1;j<num_classes;j++)
      classifiers[i][j]->adddata(idx,y,C);
  
  if(l_idx != -1){
    cl->adddata(idx,1,C);
  }
  else
    cl->adddata(idx,-1,C);
}

void IncSVM_multiclass::process_instance(int idx, int label){
  int l_idx = -1;
  for(int i=0;i<num_classes;i++){
    if((*labels)[i] == label){
      l_idx = i;
      break;
    }
  }

  if(l_idx == -1){
    cl->process_instance(idx,-1);
    return;
  }


  for(int i=0;i<num_classes;i++){
    ne[i][l_idx]++;
    ne[l_idx][i]++;
    if(i<l_idx){
      classifiers[i][l_idx]->process_instance(idx,-1);
    }
    else if(i>l_idx){
      classifiers[l_idx][i]->process_instance(idx,1);
    }
  }
  cl->process_instance(idx,1);
}

void IncSVM_multiclass::unlearn_instance(int idx, int label){
  int l_idx = -1;
  for(int i=0;i<num_classes;i++)
    if((*labels)[i] == label){
      l_idx = i;
      break;
    }

  if(l_idx == -1){
    cl->unlearn_instance(idx,-1);
    return;
  }


  for(int i=0;i<num_classes;i++){
    ne[i][l_idx]--;
    ne[l_idx][i]--;

    if(i<l_idx){
      classifiers[i][l_idx]->unlearn_instance(idx,-1);
    }
    else if(i>l_idx){
      classifiers[l_idx][i]->unlearn_instance(idx,1);
    }
  }
  cl->unlearn_instance(idx,1);
}


double IncSVM_multiclass::foreground_prob(int idx){
  return cl->svmprob(idx,1);
}

void IncSVM_multiclass::process_instance_only(int idx, int label){
  int l_idx = -1;
  for(int i=0;i<num_classes;i++){
    if((*labels)[i] == label){
      l_idx = i;
      break;
    }
  }

  if(l_idx == -1){
    cl->process_instance_only(idx,-1);
    return;
  }


  for(int i=0;i<num_classes;i++){
    ne[i][l_idx]++;
    ne[l_idx][i]++;
    if(i<l_idx){
      classifiers[i][l_idx]->process_instance_only(idx,-1);
    }
    else if(i>l_idx){
      classifiers[l_idx][i]->process_instance_only(idx,1);
    }
  }
  cl->process_instance_only(idx,1);
}

void IncSVM_multiclass::unlearn_instance_only(int idx, int label){
  int l_idx = -1;
  for(int i=0;i<num_classes;i++)
    if((*labels)[i] == label){
      l_idx = i;
      break;
    }
  
  
  if(l_idx == -1){
    cl->unlearn_instance_only(idx,-1);
    return;
  }

  for(int i=0;i<num_classes;i++){
    ne[i][l_idx]--;
    ne[l_idx][i]--;

    if(i<l_idx){
      classifiers[i][l_idx]->unlearn_instance_only(idx,-1);
    }
    else if(i>l_idx){
      classifiers[l_idx][i]->unlearn_instance_only(idx,1);
    }
  }
  cl->unlearn_instance_only(idx,1);
}


void IncSVM_multiclass::update_probabilities(int * classes){
  for(int i=0;i<num_classes;i++){
    if(classes[i] ==0)
      continue;
    for(int j=0;j<num_classes;j++)
      if(i<j)
	classifiers[i][j]->fit_sigmoid(100);
      else if(i>j)
	classifiers[j][i]->fit_sigmoid(100);
  }
  cl->fit_sigmoid(100);
}





int IncSVM_multiclass::classify(int indc){
  double probs[labels->size()];
  svmprobs(indc,probs);
  double max = -1;
  int idx = -1;
  for(unsigned int i=0;i<labels->size();i++){
    if(probs[i] > max){
      max = probs[i];
      idx = i;
    }
  }
  return (*labels)[idx];
}

void IncSVM_multiclass::estimate_probabilities1(double **r, double *probs){
  double mu[num_classes][num_classes];
  double sum = 0;
  for(int i=0;i<num_classes;i++){
    probs[i] = 0;
    for(int j=0;j<num_classes;j++)
      if(i!=j)
	probs[i] += r[i][j];
    probs[i] = probs[i] * 2/num_classes/(num_classes-1);
    sum += probs[i];
  }
  for(int i=0;i<num_classes;i++)
    probs[i] /= sum;
  
  int converged = 0;
  double prev_sum = 0;
  double epsilon = 1e-3;
  double prev_l = 0;
  int num_iter = 0;
  while(!converged && num_iter<100){
    double sum = 0;
    double l = 0;

    for(int i=0;i<num_classes;i++)
      for(int j=0;j<num_classes;j++){
	mu[i][j] = probs[i]/(probs[i] + probs[j]);
	if(i<j)
	  l += ne[i][j]*(r[i][j]*log(r[i][j]/mu[i][j])+(1-r[i][j])*log((1-r[i][j])/(1-mu[i][j])));
      }
    
    if(abs(prev_l - l)<=epsilon){
      converged = 1;
      break;
    }
    prev_l = l;
    
    for(int i=0;i<num_classes;i++){
      double r_sum = 0;
      double mu_sum = 0;
      for(int j=0;j<num_classes;j++)
	if(i!=j){
	  r_sum += r[i][j]*ne[i][j];
	  mu_sum += mu[i][j]*ne[i][j];
	}
      
      probs[i] *= r_sum/mu_sum;
      sum += probs[i];
    }
    
    for(int i=0;i<num_classes;i++)
      probs[i] /= sum;

    prev_sum = sum;
    num_iter++;
  }
}




void IncSVM_multiclass::estimate_probabilities(double **r, double *p)
{
  int k=num_classes;
	int t,j;
	int iter = 0, max_iter=max(100,k);
	double **Q=Malloc(double *,k);
	double *Qp=Malloc(double,k);
	double pQp, eps=0.005/k;
	
	for (t=0;t<k;t++)
	{
		p[t]=1.0/k;  // Valid if k = 1
		Q[t]=Malloc(double,k);
		Q[t][t]=0;
		for (j=0;j<t;j++)
		{
			Q[t][t]+=r[j][t]*r[j][t];
			Q[t][j]=Q[j][t];
		}
		for (j=t+1;j<k;j++)
		{
			Q[t][t]+=r[j][t]*r[j][t];
			Q[t][j]=-r[j][t]*r[t][j];
		}
	}
	for (iter=0;iter<max_iter;iter++)
	{
		// stopping condition, recalculate QP,pQP for numerical accuracy
		pQp=0;
		for (t=0;t<k;t++)
		{
			Qp[t]=0;
			for (j=0;j<k;j++)
				Qp[t]+=Q[t][j]*p[j];
			pQp+=p[t]*Qp[t];
		}
		double max_error=0;
		for (t=0;t<k;t++)
		{
			double error=fabs(Qp[t]-pQp);
			if (error>max_error)
				max_error=error;
		}
		if (max_error<eps) break;
		
		for (t=0;t<k;t++)
		{
			double diff=(-Qp[t]+pQp)/Q[t][t];
			p[t]+=diff;
			pQp=(pQp+diff*(diff*Q[t][t]+2*Qp[t]))/(1+diff)/(1+diff);
			for (j=0;j<k;j++)
			{
				Qp[j]=(Qp[j]+diff*Q[t][j])/(1+diff);
				p[j]/=(1+diff);
			}
		}
	}
	if (iter>=max_iter)
		printf("Exceeds max_iter in multiclass_prob\n");
	for(t=0;t<k;t++) free(Q[t]);
	free(Q);
	free(Qp);
}




void IncSVM_multiclass_cc::process_instance(int idx, int label){
  IncSVM_multiclass::process_instance(idx,label);
  
  for(int i=0;i<num_classes;i++)
    for(int j=i+1;j<num_classes;j++)
      if((*labels)[i] == label || (*labels)[j] == label){
	cc[i][j]->process_instance(idx,1);
      }
      else{
	cc[i][j]->process_instance(idx,-1);
      }
}

void IncSVM_multiclass_cc::unlearn_instance(int idx, int label){
  IncSVM_multiclass::unlearn_instance(idx,label);
  
  for(int i=0;i<num_classes;i++)
    for(int j=i+1;j<num_classes;j++)
      if((*labels)[i] == label || (*labels)[j] == label){
	cc[i][j]->unlearn_instance(idx,1);
      }
      else{
	cc[i][j]->unlearn_instance(idx,-1);
      }
}


void IncSVM_multiclass_cc::add_crossval_data(vector<int>* crossval_indices, vector<int> * crossval_labels){
  IncSVM_multiclass::add_crossval_data(crossval_indices, crossval_labels);
  for(int i=0;i<num_classes;i++)
    for(int j=i+1;j<num_classes;j++){
      vector<int> * indices = new vector<int>;
      vector<int> * c_labels = new vector<int>;
      
      for(unsigned int k=0;k<crossval_indices->size();k++){
	indices->push_back((*crossval_indices)[k]);
	if((*crossval_labels)[k] == (*labels)[i] || (*crossval_labels)[k] == (*labels)[j])
	  c_labels->push_back(1);
	else
	  c_labels->push_back(-1);
      }
      cc[i][j]->add_crossval_data(indices,c_labels);
    }
}

void IncSVM_multiclass_cc::svmprobs(int indc, double * probs){
  double **r = new double*[num_classes];
  for(int i=0;i<num_classes;i++)
    r[i] = new double[num_classes];
  
  for(int i=0;i<num_classes;i++)
    for(int j=0;j<num_classes;j++){
      if(i==j)
	continue;
      else if(i<j)
	r[i][j] = classifiers[i][j]->svmprob(indc,1) * cc[i][j]->svmprob(indc,1);
      else
	r[i][j] = classifiers[j][i]->svmprob(indc,-1) * cc[j][i]->svmprob(indc,1);
    }

  double sum = 0;
  for(int i=0;i<num_classes;i++){
    probs[i] = 0;
    for(int j=0;j<num_classes;j++)
      if(i!=j)
	probs[i] += r[i][j];
    probs[i] = probs[i] * 2/num_classes/(num_classes-1);
    sum += probs[i];
  }
  for(int i=0;i<num_classes;i++){
    probs[i] /= sum;
  }

  for(int i=0;i<num_classes;i++)
    delete r[i];
  delete r;
}


void IncSVM_multiclass_cc::adddata(int idx, int y, double C){
  IncSVM_multiclass::adddata(idx,y,C);
  for(int i=0;i<num_classes;i++)
    for(int j=i+1;j<num_classes;j++)
      cc[i][j]->adddata(idx,y,C);
}


