#ifndef _LIBSVM_H
#define _LIBSVM_H 1

#include <vector>
#include<sys/time.h>
#include<time.h>
#include <map>
#include <stdio.h>
#include <stdlib.h>

void my_assert(int assertion, char * msg);

using namespace std;

class stopwatch
{
public:
    stopwatch() : start(clock()){gettimeofday(&start1,NULL);} //start counting time
    void reset(){
      gettimeofday(&start1,NULL);
    }
    ~stopwatch(){}
    double get_time() 
        {
	  //clock_t total = clock()-start;;
	  //return double(total)/CLOCKS_PER_SEC;
	  struct timeval current;
	  gettimeofday(&current,NULL);
	  double elapsedtime = (current.tv_sec - start1.tv_sec)+(current.tv_usec-start1.tv_usec/1000000);
	  return elapsedtime;
        };
private:
    struct timeval start1;
    clock_t start;
};

class Matrix{
public:
  double **array;
  double *p;
  int size;
  int * row_size;
  int m,n;

  Matrix(){
  }
  
  Matrix(int _m, int _n){
    initialize(_m,_n);
  }

  void initialize(int _m,int _n){
    size = 0;
    m = _m;
    n = _n;
    p = new double[m*n];
    row_size = new int[m];
    array = new double*[m];
    for(int i=0;i<m;i++){
      array[i] = &p[i*n];
      row_size[i] = 0;
      for(int j=0;j<n;j++)
	array[i][j] = 0;
    }
  }

  ~Matrix(){
    printf("Deleting matrix\n");
    delete[] row_size;
    delete[] p;
    delete[] array;
  }
  
  void del(int i, int j){
    //my_assert(i<size && j<row_size[i],"Line 67: libincsvm");
    for(int k=j;k<row_size[i]-1;k++)
      array[i][k] = array[i][k+1];
    array[i][row_size[i]-1] = 0;
    row_size[i]--;
  }

  void del_col(int j){
    for(int k=0;k<size;k++)
      if(j<row_size[k])
	del(k,j);
  }

  void del_row(int row){
    for(int i=row;i<size-1;i++){
      for(int j=0;j<row_size[i];j++)
	array[i][j] = array[i+1][j];
      row_size[i] = row_size[i+1];
    }
    size--;
  }

  void print_row(int ind){
    printf("Row %d: ",ind);
    for(int i=0;i<row_size[ind];i++)
      printf("%lf ",array[ind][i]);
    printf("\n");
  }

  void push_back(int i, double val){
    array[i][row_size[i]] = val;
    row_size[i]++;
  }

  void push_back(double val){
    push_back(size-1,val);
  }

  void clear(int i){
    row_size[i] = 0;
  }
  
  void addrow(){
    row_size[size] = 0;
    size++;
    if(size >= m){
      fprintf(stderr,"Matrix (%d, %d) out of space %d\n",m,n,size);
      fflush(stderr);
      exit(-3);
    }
  }

  double * get_row(int row){
    return array[row];
  }

  int get_row_size(int row){
    return row_size[row];
  }

};



class SparseMatrix{
  double **array;
public:
  double *p;
  int size;
  int * row_size;
  int m,n;
  map<int,int> index_to_row;
  map<int,int> row_to_index;

  SparseMatrix(){
  }
  
  SparseMatrix(int _m, int _n){
    initialize(_m,_n);
  }

  void initialize(int _m,int _n){
    size = 0;
    m = _m;
    n = _n;
    p = new double[m*n];
    row_size = new int[m];
    array = new double*[m];
    for(int i=0;i<m;i++){
      array[i] = &p[i*n];
      row_size[i] = 0;
      for(int j=0;j<n;j++)
	array[i][j] = 0;
    }
  }

  ~SparseMatrix(){
    printf("Deleting matrix\n");
    delete[] row_size;
    delete[] p;
    delete[] array;
  }
  
  void del(int i, int j){
    //my_assert(i<size && j<row_size[i],"Line 67: libincsvm");
    for(int k=j;k<row_size[i]-1;k++)
      array[i][k] = array[i][k+1];
    array[i][row_size[i]-1] = 0;
    row_size[i]--;
  }

  void del_col(int j){
    for(int k=0;k<size;k++)
      if(j<row_size[k])
	del(k,j);
  }

  void del_row(int ind){
    int row = index_to_row[ind];
    for(int i=row;i<size-1;i++){
      for(int j=0;j<row_size[i];j++){
	array[i][j] = array[i+1][j];
      }
      row_to_index[i] = row_to_index[i+1];
      index_to_row[row_to_index[i]] = i;
      row_size[i] = row_size[i+1];
    }
    row_to_index[size] = -1;
    index_to_row[ind] = -1;
    size--;
  }

  void print_row(int ind){
    int row = index_to_row[ind];
    printf("Row %d: ",row);
    for(int i=0;i<row_size[row];i++)
      printf("%lf ",array[row][i]);
    printf("\n");
  }

  void push_back(int ind, double val){
    int i = index_to_row[ind];
    array[i][row_size[i]] = val;
    row_size[i]++;
  }

  void push_back(double val){
    push_back(size-1,val);
  }

  void clear(int i){
    del_row(i);
    //row_size[i] = 0;
  }

  void addrow(int ind){
    row_size[size] = 0;
    index_to_row[ind] = size;
    row_to_index[size] = ind;
    size++;
    if(size >= m){
      fprintf(stderr,"Matrix (%d, %d) out of space %d\n",m,n,size);
      fflush(stderr);
      exit(-3);
    }
  }
  
  double * get_row(int ind){
    int row = index_to_row[ind];
    return array[row];
  }

  int get_row_size(int ind){
    if(index_to_row.find(ind) != index_to_row.end() && index_to_row[ind] != -1)
      return row_size[index_to_row[ind]];
    return 0;
  }
};


/* Base incremental svm class
 */
class IncSVM{
 
 public:
  vector<double> a;
  double b;
  vector<int> ind[4];

  virtual double svmeval(int indc,std::vector<double>* Qc);

  virtual int classify(int indc){return 0;}
  
  virtual int learn(int indc,int rflag);
  
  virtual int unlearn(int indc);
  
  virtual void initialize(int m, int n);
  
  virtual void adddata(int idx, int y_new, double C_new);
  
  virtual void setY(int idx,int y_new);
  
  std::vector<double>& get_model(double& b0);
  
  //void print_vector(std::vector<int>& a);
  
  //void print_vector(std::vector<double>& a);

  virtual void process_instance(int idx, int label){}
  
  virtual void unlearn_instance(int idx, int label){}

  virtual double svmprob(int indc, int label){return 0;}
  
  virtual void svmprobs(int indc, double * probs){}

  virtual double svmprob_ispresent(vector<int>* indc, int label){return 0;}
  
  virtual void svmprobs_ispresent(vector<int>* indc, double * probs){}

  virtual void add_crossval_data(vector<int>* _crossval_indices, vector<int> * labels){}
  virtual void update_probabilities(int * classes){};

  virtual void process_instance_only(int idx, int label){}
  
  virtual void unlearn_instance_only(int idx, int label){}

  virtual void save_sigmoid(){};
  virtual void restore_sigmoid(){};
  virtual   double foreground_prob(int idx){};
  
  virtual ~IncSVM(){};
  
  double abs(double a);
  
  double get_w2();

 private:
  Matrix _Rs;
  SparseMatrix _Q;
  //Matrix _Q;
  vector<double> c;
  double deps;
  vector<double> g;
  //vector<vector<double>*> Q;
  //vector<vector<double>*> Rs;
  double scale;
  double type;
  vector<double> y;
  int perturbations;
  int max_reserve_vectors;
  int kernel_evals;
  stopwatch watch;
  

  int find(int indc, vector<int>& indices);
  
  void move_ind(int cstatus, int nstatus, int index);
  
  int move_indr(int cstatus, int indc);
  
  void bookkeeping(int indss, int cstatus, int nstatus,int& indco, int& idx);

  void select_indices(vector<double>& array, vector<int>& indices, vector<double>& res);
  
  //void matrix_star_vector(int c, vector<vector<double>*>& A,vector<double>& x,vector<double>& res);
  
  //void vector_plus_vector(int start1, int c1, vector<double>& v1, int start2, int c2, vector<double>& v2, int start_res, vector<double>& result, int size);
  
  //void vector_plus_vector(vector<int>& ind1, int c1, vector<double>& v1, int start2, int c2, vector<double>& v2, vector<int>& indres, vector<double>& result);
  
  //double vector_star_vector(vector<double>& v1, vector<double>& v2);
  
  void updateRQ(int indc);
  
  void updateRQ(vector<double>& beta, vector<double>& tmp1, int indc);
  
  void min(double * a,int size,double& res, int& ind);
  
  void min_delta(vector<int>& flags, vector<double> psi_initial, vector<double> psi_final,vector<double> psi_sens,double& min_d, int& k);
  
  void min_delta_acb(int indc,vector<double>& gamma,vector<double>& beta,int polc,int rflag, double& min_dacb, int& indss_,int& cstatus_,int& nstatus_);
  
  //void min_delta_acb(int indc,vector<double>& gamma,vector<double>& beta,int polc,double& min_delta_param,double& min_dacb, int& indss_,int& cstatus_,int& nstatus_);

};

/* Incremental svm with probabilities mapped using a sigmoid function trained on validation data */
class IncSVM_posterior : public IncSVM{
 public: 
  virtual void add_crossval_data(vector<int>* _crossval_indices, vector<int> * labels){
    crossval_indices = _crossval_indices;
    crossval_size = crossval_indices->size();
    crossval_labels = labels;
    pp = new double[crossval_size];
    double prior1=0,prior0=0;
    vector<int>& c_labels = *crossval_labels;
    for(int i=0;i<crossval_size;i++){
      if(c_labels[i]==1)
	prior1 ++;
      else if(c_labels[i]==-1)
	prior0 ++;
    }
    
    for(int i=0;i<crossval_size;i++){
      pp[i] = (prior1+1)/(prior0+prior1+2);
    }
  }


  virtual double svmprob(int indc, int label);

  virtual int classify(int indc);
  
  virtual void svmprobs(int indc, double * probs);

  virtual double svmprob_ispresent(vector<int>* indc, int label){return 0;}
  
  virtual void svmprobs_ispresent(vector<int>* indc, double * probs){}

  
  //virtual int learn(int indc,int rflag);
  
  //virtual int unlearn(int indc);

  virtual void process_instance(int idx, int label);
  
  virtual void unlearn_instance(int idx, int label);

  virtual void process_instance_only(int idx, int label);
  
  virtual void unlearn_instance_only(int idx, int label);

  void fit_sigmoid(int iter);

  virtual void save_sigmoid(){oldsigmoidA = sigmoidA;oldsigmoidB = sigmoidB;};
  virtual void restore_sigmoid(){sigmoidA = oldsigmoidA;sigmoidB = oldsigmoidB;}

  virtual double foreground_prob(int idx){};
  
  ~IncSVM_posterior(){
    delete crossval_indices;
    delete crossval_labels;
    printf("Deleted incsvm posterior\n");
  }

 protected:
  vector<int>* crossval_indices;
  vector<int>* crossval_labels;
  double * pp;
  int crossval_size;
  double sigmoidA;
  double sigmoidB;
  double oldsigmoidA;
  double oldsigmoidB;
};


/* A multi-class incremental svm */
class IncSVM_multiclass : public IncSVM_posterior{
 protected:
  IncSVM_posterior *** classifiers;
  IncSVM_posterior * cl;
  vector<int> * labels;
  int num_classes;
  void estimate_probabilities(double **r, double *probs);
  void estimate_probabilities1(double **r, double *probs);
  double ** ne;

 public:
  IncSVM_multiclass(int n, vector<int> * _labels){
    classifiers = new IncSVM_posterior**[n];
    ne = new double *[n];
    for(int i=0;i<n;i++){
      classifiers[i] = new IncSVM_posterior*[n];
      ne[i] = new double[n];
      for(int j=i+1;j<n;j++){
	classifiers[i][j] = new IncSVM_posterior();
      }
      for(int j=0;j<n;j++)
	ne[i][j] = 0;
    }
    cl = new IncSVM_posterior();
    labels = _labels;
    num_classes = n;
  }

  void initialize(int _m, int _n){
    for(int i=0;i<num_classes;i++)
      for(int j=i+1;j<num_classes;j++)
	classifiers[i][j]->initialize(_m,_n);
    cl->initialize(500,5000);
  }

  ~IncSVM_multiclass(){
    for(int i=0;i<num_classes;i++){
      for(int j=i+1;j<num_classes;j++){
	delete classifiers[i][j];
	printf("Deleted classifier %d vs %d\n",(*labels)[i],(*labels)[j]);
      }
      delete[] classifiers[i];
      printf("Deleted all classifiers for %d\n",(*labels)[i]);
    }
    delete[] classifiers;
    delete cl;
    printf("Deleted multiclass classifier\n");
  }
  
  virtual void add_crossval_data(vector<int>* indices, vector<int> * labels);

  double svmprob(int indc, int label);
  double svmprob_ispresent(vector<int>* indc, int label);
  double foreground_prob(int idx);

  virtual void svmprobs(int indc, double * probs);
  void svmprobs_ispresent(vector<int> * indc, double * probs);

  int classify(int indc);
  
  virtual void adddata(int idx, int y, double C);
    
  virtual void process_instance(int idx, int label);
  
  virtual void unlearn_instance(int idx, int label);

  virtual void process_instance_only(int idx, int label);
  
  virtual void unlearn_instance_only(int idx, int label);

  virtual void save_sigmoid(){
    for(int i=0;i<num_classes;i++)
      for(int j=i+1;j<num_classes;j++)
	classifiers[i][j]->save_sigmoid();
    cl->save_sigmoid();
  };
  virtual void restore_sigmoid(){
    for(int i=0;i<num_classes;i++)
      for(int j=i+1;j<num_classes;j++)
	classifiers[i][j]->restore_sigmoid();
    cl->restore_sigmoid();
  }


  void update_probabilities(int * classes);

};


/* A multi-class incremental svm method using a different approach for obtaining the probabilities */
class IncSVM_multiclass_cc : public IncSVM_multiclass{
 protected:
  IncSVM_posterior *** cc;

 public:
 IncSVM_multiclass_cc(int n, vector<int> * _labels) : IncSVM_multiclass(n,_labels){
    cc = new IncSVM_posterior**[n];
    for(int i=0;i<n;i++){
      cc[i] = new IncSVM_posterior*[n];
      for(int j=i+1;j<n;j++){
	cc[i][j] = new IncSVM_posterior();
      }
    }
    cl = new IncSVM_posterior();
  }

  void initialize(int _m, int _n){
    for(int i=0;i<num_classes;i++)
      for(int j=i+1;j<num_classes;j++){
	classifiers[i][j]->initialize(_m,_n);
	cc[i][j]->initialize(_m,_n);
      }
    cl->initialize(500,5000);
  }

  ~IncSVM_multiclass_cc(){
    for(int i=0;i<num_classes;i++){
      for(int j=i+1;j<num_classes;j++){
	delete cc[i][j];
	printf("Deleted classifier %d vs %d\n",(*labels)[i],(*labels)[j]);
      }
      delete[] cc[i];
      printf("Deleted classifier all classifiers for %d\n",(*labels)[i]);
    }
    delete[] cc;
    printf("Deleted multiclass classifier\n");
  }
  
  void add_crossval_data(vector<int>* indices, vector<int> * labels);

  void process_instance(int idx, int label);
  
  void unlearn_instance(int idx, int label);

  void svmprobs(int indc, double * probs);

  void adddata(int idx, int y, double C);
};



#endif
