#include "libvoi.h"
#include "libincsvm.h"
#include <math.h>
#include <fstream>

using namespace std;

ofstream effort_info_out;
extern int initial_set_size;

extern char model_file_name[1024];
extern int verbose;

void VOIActiveLearner::calculate_probabilities(int set, int size, double *pp){
  for(int i=0;i<size;i++)
    pp[i] = classifier->svmprob(data_map[set][i], (*Y)[data_map[set][i]]);
}


double VOIActiveLearner::JU(double * p, int size){
  double sum = 0;
  for(int i=0;i<size;i++)
      sum += p[i]*(1-p[i]);
  return (r[0]+r[1])*sum;
}


double VOIActiveLearner::JL(double * p, int size, int unlabeled){
  if(unlabeled)
    return JU(p,size);
  
  double sum = 0;
  for(int i=0;i<size;i++)
    sum += r[0]*(1-p[i]);
  return sum;
}


double VOIActiveLearner::calculate_jlju1(){
  double u = 0;
  int sets[] = {TRAIN, POSINSTANCES, UNLABELEDBAGS};
  int type[] = {0, 1, 1};
  
  for(int i=0;i<3;i++){
      double pp[data_map[sets[i]].size()];
      calculate_probabilities(sets[i],data_map[sets[i]].size(),pp);
      u += JL(pp,data_map[sets[i]].size(),type[i]);
  }
  return u;
}

double VOIActiveLearner::Risk(int idx, int type){
  MISparseVector * misv = (MISparseVector*)((*X)[idx]);
  if(type == 0 && misv->size() == 1){
    if(labels_index.find((*Y)[idx]) == labels_index.end())
      return 0;
    int rev_idx = labels_index[(*Y)[idx]];
    double prob = classifier->svmprob(idx, (*Y)[idx]);
    return r[rev_idx]*(1 - prob);
  }

  if(type == 0){
    double sum = 0;
    double *probs = new double[labels->size()];
    classifier->svmprobs_ispresent(&misv->indices, probs);
    
    for(unsigned int i=0;i<misv->labels.size();i++){
      if(labels_index.find(misv->labels[i]) == labels_index.end())
	continue;

      int rev_idx = labels_index[misv->labels[i]];
      sum += r[rev_idx]*(1-probs[rev_idx]);
    }
    delete probs;
    return sum;
  }

  if(type == 1){
    double sum = 0;
    double * probs = new double[labels->size()];
    classifier->svmprobs(idx,probs);
    for(unsigned int i=0;i<labels->size();i++){
      probs[i] *= classifier->foreground_prob(idx);
      sum += r[i]*(1 - probs[i])*probs[i];
    }
    delete probs;
    return sum;
  }
  
  return 0;
  double sum = 0;
  double * probs = new double[labels->size()];
  double acquired[labels->size()];
  classifier->svmprobs_ispresent(&misv->indices, probs);

  for(unsigned int i=0;i<labels->size();i++)
    acquired[i] = 0;

  for(unsigned int j=0;j<misv->labels.size();j++){
    int rev_idx = labels_index[misv->labels[j]];
    if(misv->acquired[j])
      acquired[rev_idx] = 1;
  }
  
  for(unsigned int i=0;i<labels->size();i++){
    int rev_idx = labels_index[(*labels)[i]];
    if(acquired[rev_idx])
      sum += r[rev_idx]*(1-probs[rev_idx]);
    else
      sum += r[rev_idx]*(1-probs[rev_idx])*probs[rev_idx];
  }
  delete probs;
  return sum;
}

double VOIActiveLearner::calculate_jlju(){
  double u = 0;
  int sets[] = {TRAIN, POSINSTANCES, UNLABELEDBAGS};
  int type[] = {0, 1, 2};
  double u1[3];

  for(int i=0;i<3;i++){
    for(unsigned int j=0;j<data_map[sets[i]].size();j++){
      u += Risk(data_map[sets[i]][j],type[i]);
    }
    if(i!=0)
      u1[i] = u - u1[i-1];
    else
      u1[i] = u;
  }
  return u;
}


int VOIActiveLearner::check(int set, int index){
    int flag = 1;
    for(unsigned int i=0;i<data_map[set].size();i++)
	if(data_map[set][i] == index){
	    flag = 0;
	    break;
	}
    return flag;
}

void VOIActiveLearner::delete_index(int set, int index){
    for(unsigned int i=0;i<data_map[set].size();i++)
	if(data_map[set][i] == index){
	    data_map[set].erase(data_map[set].begin()+i);
	    break;
	}
}

int VOIActiveLearner::sample_label(int index,vector<int> * labels){
  double probs[this->labels->size()];
  classifier->svmprobs(index,probs);
  double total_prob = 0;
  for(int i=0;i<labels->size();i++){
    total_prob += probs[labels_index[(*labels)[i]]];
  }

  double randnum = (double)rand()/RAND_MAX*total_prob;
  int label = 1;
  double p = 0;
  for(unsigned int i=0;i<labels->size();i++){
    int rev_idx = labels_index[(*labels)[i]];
    if(randnum >=p && randnum < p+probs[rev_idx]){
      label = (*labels)[i];
      break;
    }
    p+=probs[rev_idx];
  }
  return label;
}


/* get samples to compute expected voi using Gibbs' sampling */
int VOIActiveLearner::get_samples(vector<int> * samples, MISparseVector * misv_i){
    my_assert(misv_i->size()>=0,"Line 2069");
    int got_label[misv_i->size()];
    int num_unknown = 0;
    for(int i=0;i<misv_i->size();i++){
	got_label[i] = 1-check(TRAIN,misv_i->getInstance(i)->index);
	num_unknown += 1 - got_label[i];
    }
    int max_samples = sample_size;
    if(num_unknown == 0)
	return 0;
    for(int i=0;i<misv_i->size();i++){
	int index = misv_i->getInstance(i)->index;
	my_assert(index < (int)Y->size(),"Line 2083");
	if(got_label[i])
	    samples[i].push_back((*Y)[index]);
	else{
	  vector<int> * lbl = labels;
	  if(misv_i->labels.size() != 0)
	    lbl = &misv_i->labels;
	  int label = sample_label(index,lbl);
	  samples[i].push_back(label);
	}
    }
    stopwatch swa;
    classifier->save_sigmoid();
    for(int samp=1;samp<max_samples;samp++){
      fflush(stdout);
      stopwatch sw_samp;
      for(int inst = 0;inst<misv_i->size();inst++){
	stopwatch sw;
	int index = misv_i->getInstance(inst)->index;
	my_assert(index < (int)Y->size(),"Line 2102");
	fflush(stdout);
	if(got_label[inst]){
	  samples[inst].push_back((*Y)[misv_i->getInstance(inst)->index]);
	  continue;
	}
	int flags[labels->size()];
	for(int j=0;j<labels->size();j++)
	  flags[j] = 0;
	stopwatch sw1;
	for(int j=0;j<misv_i->size();j++){
	  if(j==inst || got_label[j])
	    continue;
	  my_assert(samples[j].size()>0,"Line 2116");
	  sw1.reset();
	  classifier->process_instance_only(misv_i->getInstance(j)->index,samples[j][samples[j].size()-1]);
	  if(samples[j][samples[j].size()-1]!=1)
	    flags[labels_index[samples[j][samples[j].size()-1]]] = 1;
	}
	sw1.reset();
	classifier->update_probabilities(flags);
	sw1.reset();

	vector<int> * lbl = labels;
	if(misv_i->labels.size() != 0)
	  lbl = &misv_i->labels;
	int label = sample_label(index,lbl);
	samples[inst].push_back(label);
	sw1.reset();
	for(int j=misv_i->size()-1;j>=0;j--){
	  if(j==inst || got_label[j])
	    continue;
	  int index = misv_i->getInstance(j)->index;
	  my_assert(index < (int)Y->size(),"Line 2145");
	  classifier->unlearn_instance_only(misv_i->getInstance(j)->index,samples[j][samples[j].size()-1]);//,(*Y)[misv_i->getInstance(j)->index]);
	}
	classifier->restore_sigmoid();
	sw1.reset();
      }
    }
    if(verbose == 1)
    printf("Time to get samples %lf\n",swa.get_time());
    return num_unknown;
}

/* compute voi for obtaining labels on all instances in a bag and select the bag with largest voi */
void VOIActiveLearner::select_bag(int& best, double& voi, double& cost){
    best = -1;
    voi = -1e20;
    double u = calculate_jlju();
    double max = -1e20;

    classifier->save_sigmoid();
    for(unsigned int i=0;i<data_map[TRAINPOS].size();i++){
      stopwatch sw;
      MISparseVector * misv_i = (MISparseVector*)(*X)[data_map[TRAINPOS][i]];
      if(verbose == 1)
	printf("Processing instance %d size %d\n",i,misv_i->size());
      fflush(stdout);
      if(misv_i->size()<1)
	    continue;
	vector<int> samples[misv_i->size()];
	double u1 = 0;
	double u2 = 0;
	double voi;
	get_samples(samples,misv_i);
	
	if(samples[0].size()==0)
	    continue;

	fflush(stdout);
	for(unsigned int samp=sample_start;samp<samples[0].size();samp++){
	    int num_added = 0;
	    vector<int> labels;
	    vector<int> indices;

	    int flags[this->labels->size()];
	    for(int j=0;j<this->labels->size();j++)
	      flags[j] = 0;
	    for(int inst = 0;inst<misv_i->size();inst++){
	      int index = misv_i->getInstance(inst)->index;
		my_assert(index < (int)Y->size(),"Line 2188");
		if(check(TRAIN,index)){
		    data_map[TRAIN].push_back(index);
		    delete_index(POSINSTANCES,index);
		    labels.push_back((*Y)[index]);
		    indices.push_back(inst);
		    (*Y)[index] = samples[inst][samp];
		    classifier->process_instance_only(index,samples[inst][samp]);
		    flags[labels_index[samples[inst][samp]]] = 1;
		    num_added++;
		}
	    }
	    
	    classifier->update_probabilities(flags);
	    double a = calculate_jlju();
	    u1 += a;

	    for(int f = indices.size()-1;f>=0;f--){
	      int inst = indices[f];
		int idx = data_map[TRAIN].back();
		my_assert(idx < (int)Y->size(),"Line 2231");
		my_assert(labels.size()>0,"Line 2232");
		(*Y)[idx] = labels.back();
		labels.pop_back();
		data_map[POSINSTANCES].push_back(data_map[TRAIN].back());
		data_map[TRAIN].pop_back();
		classifier->unlearn_instance_only(idx,samples[inst][samp]);//(*Y)[idx]);
	    }
	    classifier->restore_sigmoid();
	}
	
	u1 /= (samples[0].size()-sample_start);
	voi = u - u1 - misv_i->pcost;
	if(voi>max){
	  max = voi;
	  best = i;
	  cost = misv_i->acost;
	}
	
	double oldu = calculate_jlju();
	int num_added = 0;
	
	for(int inst = 0;inst<misv_i->size();inst++){
	  int index = misv_i->getInstance(inst)->index;
	  my_assert(index < (int)Y->size(),"Line 2188");
	  if(check(TRAIN,index)){
	    data_map[TRAIN].push_back(index);
	    delete_index(POSINSTANCES,index);
	    classifier->process_instance(index,(*Y)[index]);
	    num_added++;
	  }
	}
	
	double actvoi = oldu - calculate_jlju() - misv_i->pcost;
		
	for(int f = 0;f<num_added;f++){
	  int idx = data_map[TRAIN].back();
	  my_assert(idx < (int)Y->size(),"Line 2231");
	  data_map[POSINSTANCES].push_back(data_map[TRAIN].back());
	  data_map[TRAIN].pop_back();
	  classifier->unlearn_instance(idx,(*Y)[idx]);
	}
	
	if(verbose == 1)
	  printf("Set %d Instance %d Voi %lf Actual VOi %lf num_added %d max %lf u %lf u1 %lf out %lf cost %lf time %f\n", TRAINPOS, data_map[TRAINPOS][i], voi, actvoi,num_added, max, u, u1,classifier->svmeval(data_map[TRAINPOS][i],NULL),misv_i->acost, sw.get_time());
	fflush(stdout);
    }
    voi = max;
}

/* compute voi for all instance labels and select the example with largest voi */
double VOIActiveLearner::select_instance(int Set, int& best, double& voi, double& cost){
    if(Set==TRAINPOS){
      select_bag(best,voi,cost);
      return 0;
    }

    if(Set == UNLABELEDBAGS)
      return select_unlabeled_bag(Set,best,voi,cost);

    double max = -1e20;
    double cost_of_labeling = cost_instance;
    best = -1;

    double u = calculate_jlju();
    
    for(int i=0;i<(int)data_map[Set].size();i++){
	int index = data_map[Set][i];
	double fprob = classifier->foreground_prob(index);
	int label = (*Y)[index];
	MISparseVector * misv_i = (*bag_pointer)[index];
	double u1 = 0;
	double voi;
	data_map[Set].erase(data_map[Set].begin()+i);
	data_map[TRAIN].push_back(index);
	
	stopwatch sw;
	double * probs = new double[labels->size()];
	classifier->svmprobs(index, probs);

	double prob1 = 0;
	
	if(fprob < .2){
	  printf("Skipping instance %d label %d\n",index,label);
	  continue;
	}
	
	vector<int> * lbl = labels;
	if(misv_i->labels.size() != 0)
	  lbl = &misv_i->labels;
	
	for(unsigned int j=0;j<lbl->size()/*misv_i->labels.size()*/;j++){
	  int xlabel = (*lbl)[j];
	  (*Y)[index] = xlabel;
	  classifier->process_instance(index,xlabel);
	  double tmpu = calculate_jlju();
	  classifier->unlearn_instance(index,xlabel);
	  double v =  u - tmpu - cost_of_labeling;
	  if(v<0)
	    v = 0;
	  probs[labels_index[xlabel]] *= fprob;
	  u1 += tmpu * probs[labels_index[xlabel]];
	  prob1 += probs[labels_index[xlabel]];
	}
	u1 += (1-prob1)*u;
	delete probs;
	(*Y)[index] = label;

	classifier->process_instance(index,label);
	double actu1 = calculate_jlju();
	classifier->unlearn_instance(index,label);
	double actvoi = u - actu1 - misv_i->pcost/misv_i->size();

	data_map[Set].insert(data_map[Set].begin()+i,index);
	data_map[TRAIN].pop_back();

	voi = u - u1 - misv_i->pcost/misv_i->size();
	if(voi>max){
	    max = voi;
	    best = i;
	    cost = misv_i->acost/misv_i->size();
	}
	if(verbose == 1)
	  printf("Set %d Instance %d label %d Voi %lf actual voi %lf max %lf acost %lf pcost %lf time: %f prob: %f\n",Set,i,(*Y)[index],voi,actvoi,max,misv_i->acost,misv_i->pcost,sw.get_time(),fprob);
    }
    voi = max;
    return voi;
}


void reset_kernel_row(int i);

/* compute voi for bag level labels and select the bag with largest voi */
double VOIActiveLearner::select_unlabeled_bag(int set, int& best, double& voi, double& cost){
    double max = -1e20;
    double cost_of_labeling = cost_unlabeled_bag;
    double u = calculate_jlju();
    double bestjlju = -1;
    best = -1;
    
    for(unsigned int i=0;i<data_map[set].size();i++){
	int index = data_map[set][i];
	MISparseVector * x = (MISparseVector*)((*X)[index]);
	double u1 = 0;
	double voi;
	double u0[2];
	double total_prob = 0;

	stopwatch sw;
	for(unsigned int j=0;j<labels->size();j++){
	  int label = (*labels)[j];
	  int lidx = -1;
	  for(unsigned int k=0;k<x->labels.size();k++)
	    if(x->labels[k] == label){
	      lidx = k;
	      break;
	    }
	  if(lidx != -1 && x->acquired[lidx])
	    continue;
	  
	  int flag = 1;
	  if(lidx != -1)
	    x->acquired[lidx] = 1;
	  else{
	    flag = 0;
	    x->labels.push_back(label);
	    x->acquired.push_back(1);
	  }

	  double prob = classifier->svmprob_ispresent(&x->indices,(*labels)[j]); 	  
	  double sum = 0;
	  for(int k=0;k<x->size();k++){
	    x->probs[k] = classifier->svmprob(x->indices[k],label);
	    sum += x->probs[k];
	  }
	  for(int k=0;k<x->size();k++)
	    x->probs[k] = x->probs[k] /sum * x->size();

	  reset_kernel_row(index);
	  classifier->process_instance(index,(*labels)[j]);
	  double tmpu = calculate_jlju();
	  classifier->unlearn_instance(index,(*labels)[j]);
	  reset_kernel_row(index);

	  for(int k=0;k<x->size();k++){
	    x->probs[k] = 1;
	  }
	  if(flag)
	    x->acquired[lidx] = 0;
	  else{
	    x->labels.pop_back();
	    x->acquired.pop_back();
	  }

	  u1 += tmpu * prob;
	  total_prob += prob;
	}
	
	u1 /= total_prob;
	voi = u - u1 - cost_of_labeling;
	if(verbose == 1)
	  printf("processed instance %d of size %d with voi %lf time %f\n",i,data_map[set].size(),voi,sw.get_time());
	fflush(stdout);
    	if(voi>max){
	    max = voi;
	    best = i;
	    bestjlju = u1;
	}
    }
    voi = max;
    cost = cost_of_labeling;
    return bestjlju;
}

void add_kernel_row();


/* adds selected examples into the training set and retrains the classifier 
 */
int VOIActiveLearner::add_new_data(int set, int i){
  if(set==UNLABELEDBAGS){
	int index = data_map[set][i];
	MISparseVector * x = (MISparseVector *)((*X)[index]);
	int lidx = -1;
	for(unsigned int k=0;k<x->labels.size();k++)
	  if(!x->acquired[k]){
	    lidx = k;
	    break;
	  }
	
	if(lidx == -1){
	  data_map[UNLABELEDBAGS].erase(data_map[UNLABELEDBAGS].begin()+i);
	  return -1;
	}
	double sum = 0;
	for(int k=0;k<x->size();k++){
	  x->probs[k] = classifier->svmprob(x->indices[k],x->labels[lidx])*x->size();
	  sum += x->probs[k];
	}
	
	for(int k=0;k<x->size();k++){
	  x->probs[k] = x->probs[k]/sum*x->size();
	}

	MISparseVector * instance = new MISparseVector;
	instance->addInstance(x->getRepresentation());
		
	for(int k=0;k<x->size();k++)
	  x->probs[k] = 1;
	X->push_back(instance);
	Y->push_back(x->labels[lidx]);
	add_kernel_row();
	int y = (*Y)[X->size()-1];
	double c = C * instance->norm();
	classifier->adddata(X->size()-1,y,c);
	classifier->process_instance(X->size()-1,y);
	x->acquired[lidx] = 1;
	if(lidx == (int)x->labels.size()-1){
	  data_map[UNLABELEDBAGS].erase(data_map[UNLABELEDBAGS].begin()+i);
	  data_map[TRAIN].push_back(index);
	}
	return y;
    }
    
    if(set==POSINSTANCES){
	int index = data_map[set][i];
	data_map[set].erase(data_map[set].begin()+i);
	data_map[TRAIN].push_back(index);
	classifier->process_instance(index,(*Y)[index]);
	return (*Y)[index];

	int bagidx = -1;
	for(unsigned int i=0;i<X->size();i++)
	  if((*bag_pointer)[index]==(*X)[i]){
		bagidx = i;
		break;
	    }
	if(bagidx !=-1){
	  classifier->unlearn_instance(bagidx,(*Y)[bagidx]);
	  (*bag_pointer)[index]->deletePointer((*X)[index]);
	  classifier->process_instance(bagidx,(*Y)[bagidx]);
	}
    }
	
    if(set==TRAINPOS){
	MISparseVector *misv_i = (MISparseVector *) (*X)[data_map[TRAINPOS][i]];
	for(int k=0;k<misv_i->size();k++){
	    lasvm_sparsevector_t * bag_instance = misv_i->getInstance(k);
	    for(unsigned int iter = 0;iter<data_map[POSINSTANCES].size();iter++)
		if(data_map[POSINSTANCES][iter] == bag_instance->index){
		    data_map[POSINSTANCES].erase(data_map[POSINSTANCES].begin()+iter);
		    break;
		}
	    if(check(TRAIN,bag_instance->index)){
	      classifier->process_instance(bag_instance->index,(*Y)[bag_instance->index]);
	      if(labels_index.find((*Y)[bag_instance->index]) != labels_index.end())
		data_map[TRAIN].push_back(bag_instance->index);
	      if(verbose == 1)
		printf("Adding new data %d, %d to the classifier\n",bag_instance->index,(*Y)[bag_instance->index]);
	    }
	}
	int index = data_map[TRAINPOS][i];
	delete_index(TRAIN,index);
	data_map[TRAINPOS].erase(data_map[TRAINPOS].begin()+i);
    }
    return -1;
}

/* Main active/random selection function. Selects and adds a single example and updates the classifier */
double VOIActiveLearner::run_once(int& _index,int& _set, int& _label, char * imgname,int& _actual){
  char showsel_filename[200];
  sprintf(showsel_filename, "%s-showsel", model_file_name);
  ofstream showsel;
  
  if(iteration == 0)
    showsel.open(showsel_filename, ios::out);
  else
    showsel.open(showsel_filename, ios::app);
  
  iteration++;

  char einfo[200];
  sprintf(einfo, "effort_info%d.out", initial_set_size);
  effort_info_out.open(einfo,ios::out);

  vector<int> best;
  vector<double> voi;
  vector<double> costs;
  vector<int> sets;
  int totalsets[] = {TRAINPOS, UNLABELEDBAGS, POSINSTANCES};
  for(int i=0;i<3;i++){
    if(data_map[totalsets[i]].size()!=0){
      best.push_back(-1);
      voi.push_back(-1);
      costs.push_back(-1);
      sets.push_back(totalsets[i]);
    }
  }
      
  if(sets.size()==0){
    printf("INFO: empty sets\n");
    return -1;
  }
  int best_set = -1;
  double best_voi = -1e20;
  int best_index = -1;
  double best_cost = -1;
      
  if(mi_selection_type == ACTIVE_ALL){
    for(unsigned int i=0;i<sets.size();i++){
      select_instance(sets[i],best[i],voi[i],costs[i]);
      printf("Best instance from set %d is %d, %d with voi %lf\n",sets[i], best[i], data_map[sets[i]][best[i]],voi[i]);
      if(voi[i]>best_voi){
	best_voi = voi[i];
	best_set = sets[i];
	best_index = best[i];
	best_cost = costs[i];
      }
    }
  }
  else if(mi_selection_type == ACTIVE_INSTANCE){
    best_set = POSINSTANCES;
    select_instance(POSINSTANCES,best_index,best_voi,best_cost);
    printf("Best instance from set %d is %d with voi %lf\n",POSINSTANCES, best_index, best_voi);
  }
  else if(mi_selection_type == ACTIVE_BAG){
    best_set = UNLABELEDBAGS;
    select_unlabeled_bag(UNLABELEDBAGS,best_index,best_voi,best_cost);
    printf("Best instance from set %d is %d with voi %lf\n",UNLABELEDBAGS, best_index, best_voi);
  }
  else if(mi_selection_type == RANDOM_ALL){
    double randnum = (double)rand()/RAND_MAX;
    int i = -1;
    double p = 0;
    int total_size = 0;
    for(unsigned int j=0;j<sets.size();j++)
      total_size += data_map[sets[j]].size();
    for(unsigned int j=0;j<sets.size();j++){
      if(randnum >= p && randnum < p + (double)data_map[sets[j]].size()/total_size){
	i = j;
	break;
      }
      p += (double)data_map[sets[j]].size()/total_size;
    }
    printf("Set is %d\n",sets[i]);
    my_assert(i!=-1,"Line 2733");
    best_set = sets[i];
    //best_set = TRAINPOS;
    if(data_map[best_set].size() != 0)
      best_index = (int)floor((double)rand()/RAND_MAX*data_map[best_set].size());
    else
      best_index = -1;
    if(best_set == UNLABELEDBAGS)
      best_cost = cost_unlabeled_bag;
    else if(best_set == TRAINPOS)
      best_cost = ((MISparseVector*)(*X)[data_map[best_set][best_index]])->acost;
    else
      best_cost = ((MISparseVector*)(*X)[data_map[best_set][best_index]])->acost/((MISparseVector*)(*X)[data_map[best_set][best_index]])->size();
      
  }
  else if(mi_selection_type == RANDOM_INSTANCE){
    best_set = POSINSTANCES;
    if(data_map[best_set].size() != 0)
      best_index = (int)floor((double)rand()/RAND_MAX*data_map[best_set].size());
    else
      best_index = -1;
    best_cost = ((MISparseVector*)(*X)[data_map[best_set][best_index]])->acost/((MISparseVector*)(*X)[data_map[best_set][best_index]])->size();
  }
  else if(mi_selection_type == RANDOM_BAG){
    best_set = UNLABELEDBAGS;
    if(data_map[best_set].size() != 0)
      best_index = (int)floor((double)rand()/RAND_MAX*data_map[best_set].size());
    else
      best_index = -1;
    best_cost = cost_unlabeled_bag;
  }
  
  effort_info_out.close();
     
  if(best_index == -1)
    return -1;

  double oldu = calculate_jlju();
  _index = best_index;
  _set = best_set;

  printf("Best index is %d %d\n",best_index,data_map[best_set][best_index]);
  printf("Adding new data TRAIN:%d POSINSTANCES:%d TRAINPOS:%d UNLABELED:%d\n",data_map[TRAIN].size(),data_map[POSINSTANCES].size(),data_map[TRAINPOS].size(),data_map[UNLABELEDBAGS].size());
  printf("Callign add data\n");
  _label = add_new_data(best_set,best_index);
  double actual_voi = oldu - calculate_jlju() - best_cost;
  printf("Actual value of VOI for index %d is %lf\n",best_index,actual_voi);
  return best_cost;
}

/* Trains the classifier on the initial set */
void VOIActiveLearner::do_initial_training(){
  int set = TRAIN;
  for(unsigned int i=0;i<data_map[set].size();i++){
    classifier->process_instance(data_map[set][i],(*Y)[data_map[set][i]]);
  }
}
