#ifndef _LIBVOI_H
#define _LIBVOI_H 1

#define TRAIN 0
#define TEST  1
#define UNIVERSUM 2
#define UNLABELED 3
#define SV 4
#define POSINSTANCES 5
#define UNLABELEDBAGS 6
#define TRAINPOS 7
#define CROSSVAL 8
#define TMP 9

#define ACTIVE_ALL 0
#define ACTIVE_INSTANCE 2
#define ACTIVE_BAG 4
#define RANDOM_ALL 1
#define RANDOM_INSTANCE 3
#define RANDOM_BAG 5
#define ACTIVE_INSTANCE_ACT 6
#define ACTIVE_BAG_ACT 8
#define ACTIVE_ALL_ACT 10
#define ACTIVE_ALL_INST_ACT 12

#define SEQUENTIAL 10

#include "MISparseVector.h"
#include "libincsvm.h"
#include <map>
#include <string.h>
#include <iostream>

using namespace std;

class ActiveLearner{
 public:
  double run_once(int&,int&,int&,char*,int&){return 0;};
  
  ActiveLearner(){};
  ActiveLearner(IncSVM * _classifier, vector<lasvm_sparsevector_t*> * _X, vector<int> * _Y, vector<int>* _data_map, vector<MISparseVector *>* _bag_pointer,char * _act_file){
    classifier = _classifier;
    X = _X;
    Y = _Y;
    data_map = _data_map;
    bag_pointer = _bag_pointer;
    strcpy(act_file, _act_file);
    ifstream acc_in(act_file,ios::in);
    if(acc_in.is_open()){
      while(!acc_in.eof()){
	if(acc_in.peek()==EOF)
	  break;
	int tmp1,tmp2;
	double tmp3;
	acc_in >> tmp1 >> tmp2 >>tmp3;
	if(!acc_in.good())
	  break;
	selected_set.push_back(tmp1);
	selected_indices.push_back(tmp2);
	selected_cost.push_back(tmp3);
      }
    }
    iteration = 0;
    acc_in.close();
  }

  void add_classifier(IncSVM * _classifier){
    classifier = _classifier;
  }

  void add_data(vector<lasvm_sparsevector_t*> * _X, vector<int> * _Y, vector<int> * _data_map){
    X = _X;
    Y = _Y;
    data_map = _data_map;
  }

  void add_labels(vector<int> *_labels){
    labels = _labels;
    for(unsigned int i=0;i<labels->size();i++)
      labels_index[(*labels)[i]] = i;
  }
  
 protected:
  IncSVM * classifier;
  vector<MISparseVector *> * bag_pointer;
  vector <lasvm_sparsevector_t*> * X; // feature vectors
  vector <int> *Y;                   // labels
  vector <int> * data_map;         // maps to the indices of train/test/... data
  vector<int> * labels;
  map<int,int> labels_index;
  char act_file[200];
  vector<int> selected_set;
  vector<int> selected_indices;
  vector<double> selected_cost;
  int iteration;
};


class VOIActiveLearner : public ActiveLearner{
 public:
  double run_once(int&, int&,int&,char*,int&);
  void set_parameters(double _C, double _r12, double _cost_instance, double _cost_positive_bag, double _cost_unlabeled_bag, int selection_type){
    C = _C;
    r = new double[labels->size()];
    for(unsigned int i=0;i<labels->size();i++)
      r[i] = _r12;
    cost_instance = _cost_instance;
    cost_positive_bag = _cost_positive_bag;
    cost_unlabeled_bag = _cost_unlabeled_bag;
    sample_size = 50;
    sample_start = 25;
    mi_selection_type = selection_type;
  }
  
  void set_sampling_parameters(int _sample_size, int _sample_start){
    sample_size = _sample_size;
    sample_start = _sample_start;
  }
  
  void do_initial_training();
  
  VOIActiveLearner(IncSVM * _classifier, vector<lasvm_sparsevector_t*> * _X, vector<int> * _Y, vector<int>* _data_map, vector<MISparseVector *> * _bag_pointer,char* act_file) : ActiveLearner(_classifier,_X, _Y, _data_map,_bag_pointer,act_file){
  }
  int add_new_data(int set, int i);  
 private:
  double *r;
  //double r21;
  double cost_instance;
  double cost_positive_bag;
  double cost_unlabeled_bag;
  double C;
  int sample_size;
  int sample_start;
  int mi_selection_type;

  double JU(double * p, int size);
  double JL(double * p, int size, int unlabeled);
  double calculate_jlju();
  double calculate_jlju1();
  void calculate_probabilities(int set, int size, double *pp);
  int get_samples(vector<int> * samples, MISparseVector * misv_i);
  int sample_label(int index,vector<int> *);
  void select_bag(int& best, double& voi, double& cost);
  void select_bag_act(int& best, double& voi, double& cost);
  double select_instance(int set, int& best, double& voi, double& cost);
  double select_instance_act(int set, int& best, double& voi, double& cost);
  double select_instance_acc(int set, int& best, double& voi, double& cost);
  int check(int set, int index);
  void delete_index(int set, int index);
  double Risk(int idx, int type);
  double select_unlabeled_bag(int set, int& best, double& voi, double& cost);
  double select_unlabeled_bag_act(int set, int& best, double& voi, double& cost);
};






























#endif
