#include <assert.h>
#include "strutil.h"

void GetKeyValue(const string& key_value, string& key, string& value) {
  string::size_type pos = key_value.find(' ');
  assert(pos != string::npos);

  key = key_value.substr(0, pos);
  value = key_value.substr(pos + 1);
}


bool IsIncluded(char c, const char* separators) {
  int i = 0;
  while (separators[i] != '\0') {
    if (separators[i] == c)
      return true;
    i++;
  }

  return false;
}


void SplitStringUsing(const string& value,
		      const char* separators,
		      vector<string>* tokens) {
  unsigned int i = 0;
  unsigned int begin = 0;
  while (i < value.size()) {
    if (IsIncluded(value[i], separators)) {
      if (i > begin) {
	tokens->push_back(value.substr(begin, i - begin));
	begin = i;
      }
      begin++;
      i++;
    }
    else
      i++;
  }

  // Add last segment
  if (begin < value.size())
    tokens->push_back(value.substr(begin));
}


void TrimStringLeft(string* value,
		    const char* separators) {
  unsigned int i = 0;
  while (i < value->size()) {
    if (!IsIncluded((*value)[i], separators))
      break;
    i++;
  }

  *value = value->substr(i);
}


void TrimStringRight(string* value,
		     const char* separators) {
  unsigned int i = 0;
  while (i < value->size()) {
    if (!IsIncluded((*value)[value->size() - i - 1], separators))
      break;
    i++;
  }

  *value = value->substr(0, value->size() - i);
}


void StripWhiteSpace(string* value) {
  TrimStringLeft(value, " ");
  TrimStringRight(value, " ");
}
